package btactics

import edu.cmu.cs.ls.keymaerax.bellerophon._
import edu.cmu.cs.ls.keymaerax.btactics.PolynomialArith._
import edu.cmu.cs.ls.keymaerax.btactics.TactixLibrary._
import edu.cmu.cs.ls.keymaerax.btactics._
import edu.cmu.cs.ls.keymaerax.core._
import edu.cmu.cs.ls.keymaerax.parser.StringConverter._
import testHelper.KeYmaeraXTestTags.{IgnoreInBuildTest, SlowTest}

import scala.collection.immutable._

/**
  * Invariant arithmetic examples
  * From Lectures 10/11
  * Created by yongkiat on 2/3/16.
  */
@edu.cmu.cs.ls.keymaerax.tags.SlowTest
class InvariantArithTests extends TacticTestBase {


  "InvariantArith" should "solve Lecture 10 DI arithmetic" in withMathematica { qeTool =>
    //All of these prove by normalizing both sides to 0 = 0
    //L10 Example 1
    val i1 = "v^2 + w^2 = r()^2 -> [{v'=w,w'=-v}] v^2+w^2 = r()^2"
    //L10 Example 11
    val i2 = "x^2+x^3-y^2 - c() = 0 -> [{x'=-2*y,y'=-2*x-3*x^2}] x^2+x^3-y^2 - c() = 0"
    //L10 Example 12
    val i3 = "x^4*y^2 + x^2*y^4-3*x^2*y^2+1= c()-> [{x'=2*x^4*y+4*x^2*y^3-6*x^2*y , y'=-4*x^3*y^2-2*x*y^4+6*x*y^2}]x^4*y^2 + x^2*y^4-3*x^2*y^2+1=c()"
    val i4 = "x^4+y^4 = 0 -> [{x'=4*y^3,y'=-4*x^3}]x^4+y^4 = 0"
    //L10 Exercises (6,7,8,9)
    val i5 = "x*y=c() -> [{x'=-x,y'=y,z'=-z}]x*y=c()"
    val i6 = "x^2+4*x*y-2*y^3 -y =1 -> [{x'=-1+4*x-6*y^2,y'=-2*x-4*y}]x^2+4*x*y-2*y^3 -y =1 "
    val i7 = "x^2+y^3/3 = c() -> [{x'=y^2,y'=-2*x}]x^2+y^3/3 = c()"
    val i8 = "(u^2 + v^2 + A()*x^2+B()*y^2)/2 + x^2*y - 1/3 * eps()*y^3 = 0 -> [{x'=u,y'=v,u'=-A()*x-2*x*y,v'=-B()*y+eps()*y^2-x^2}](u^2 + v^2 + A()*x^2+B()*y^2)/2 + x^2*y - 1/3 *eps()*y^3 = 0 "

    val problems = List(i1,i2,i3,i4,i5,i6,i7,i8).map(_.asFormula)

    val tactic = implyR(1) & dI('diffInd)(1) < (QE, //QE just for reflexivity, no problem
      chase(1) &
        normaliseAt(1, 0 :: Nil) &
        normaliseAt(1, 1 :: Nil) & cohideR(1) & byUS("= reflexive")
      )

    val res = problems.map(proveBy(_, tactic))
    res.forall(_.isProved) shouldBe true
  }

  "InvariantArith" should "solve Lecture 11 DI arithmetic (fig 11.5)" in withMathematica { qeTool =>
    val i = "1/3 <= 5 * x^2 -> [{x'=x^3}] 1/3 <= 5 * x^2".asFormula
    //No linear elimination
    val witness = List(("10","wit__0*x^2")).map( s => (s._1.asTerm,s._2.asTerm))
    // The first inequation is used here
    val ineq = List(0)
    val tactic = implyR(1) & dI('diffInd)(1) < (QE, //QE just for reflexivity, no problem
      chase(1) &
        DebuggingTactics.print("Before normalizing") &  prop &
        OnAll(prepareArith2) &
        printGoal &
        genWitnessTac(ineq,witness)
      )
    val res = proveBy(i, tactic)
    res shouldBe 'proved
  }

  "InvariantArith" should "solve Lecture 11 odd order dynamics (Example 11.4)" in withMathematica { qeTool =>
    val i = "x^2 >= 2 -> [{x'=x^5+7*x^3+2*x}] x^2 >= 2".asFormula
    val witness = List(("2","wit__0*x^3 + 7/5*wit__0*x"),("42/5","wit__0*x^2"),("2/25","wit__0*x")).map( s => (s._1.asTerm,s._2.asTerm))
    val ineq = List(0)
    val tactic = implyR(1) & dI('diffInd)(1) < (QE, //QE just for reflexivity, no problem
      chase(1) &
        DebuggingTactics.print("Before normalizing") &  prop &
        OnAll(prepareArith2) &
        printGoal &
        genWitnessTac(ineq,witness)
      )
    val res = proveBy(i, tactic)
    res shouldBe 'proved
  }

  "InvariantArith" should "solve Lecture 11 even order dynamics (Example 11.5)" in withMathematica { qeTool =>
    val i = "x^3 >= 2 -> [{x'=x^4+6*x^2+5}] x^3 >= 2".asFormula
    val witness = List(("3","wit__0*x^3 + 20/9*wit__0*x"),("14/3","wit__0*x^2"),("5/27","wit__0*x")).map( s => (s._1.asTerm,s._2.asTerm))
    val ineq = List(0)
    val tactic = implyR(1) & dI('diffInd)(1) < (QE, //QE just for reflexivity, no problem
      chase(1) &
        DebuggingTactics.print("Before normalizing") &  prop &
        OnAll(prepareArith2) &
        printGoal &
        genWitnessTac(ineq,witness)
      )
    val res = proveBy(i, tactic)
    res shouldBe 'proved
  }

  "InvariantArith" should "solve Lecture 11 DI arithmetic (conjuncts)" in withMathematica { qeTool =>
    val i = "v^2 + w^2 <= r()^2 & v^2 + w^2 >= r()^2 -> [{v'=w,w'=-v}] (v^2 + w^2 <= r()^2 & v^2 + w^2 >= r()^2)".asFormula

    //val insts = List((0,"1".asTerm),(1,"3*wit__0^2*x^2".asTerm))

    val tactic = implyR(1) & DebuggingTactics.print("Entering diff ind") & dI('diffInd)(1) < (QE, //QE just for reflexivity, no problem
      DebuggingTactics.print("Before chase") &
        chase(1) &
        DebuggingTactics.print("Before normalizing") &  prop &
        OnAll(prepareArith2) &
        printGoal &
        OnAll(genWitnessTac(List(0),List())) // Trivial because the inequation says 0 = 0
      )

    val res = proveBy(i, tactic)
    res shouldBe 'proved
  }

  "InvariantArith" should "solve Lecture 11 DI quartic dynamics (Example 11.7)" in withMathematica { qeTool =>
    val i = "x^3 >= -1 & a() >= 0 -> [{x'=(x-3)^4 +a()}] x^3 >= -1".asFormula

    val ineq = List(0)
    val linear = List((1,"a()","wit__1^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm))
    val witness = List(("3"," wit__0*x^3 - 6*wit__0*x^2 + 9*wit__0*x"),("3","wit__0*wit__1*x")).map( s => (s._1.asTerm,s._2.asTerm))
    //val insts = List((0,"1".asTerm),(1,"3*wit__0^2*x^2".asTerm))

    val tactic = implyR(1) & DebuggingTactics.print("Entering diff ind") & dI('diffInd)(1) < (QE, //QE just for reflexivity, no problem
      DebuggingTactics.print("Before chase") &
        chase(1) &
        DebuggingTactics.print("Before normalizing") &  prop &
        OnAll(prepareArith2) &
        printGoal & linearElim(linear) & genWitnessTac(ineq,witness)
      )

    val res = proveBy(i, tactic)
    res shouldBe 'proved
  }

  "InvariantArith" should "solve Lecture 11 DI damped oscillator" in withMathematica { qeTool =>
    val i = "w()^2*x^2 + y^2 <= c()^2 -> [{x'=y,y'=-w()^2*x-2*d()*w()*y & (w()>=0 & d()>=0)}]w()^2*x^2 + y^2 <= c()^2".asFormula

    val ineq = List(0)
    val linear = List((1,"d()","wit__1^2","1"),(1,"w()","wit__2^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm))
    val witness = List(("4","wit__0*y*wit__3*wit__4")).map( s => (s._1.asTerm,s._2.asTerm))

    val tactic = implyR(1) & dI('diffInd)(1) < (QE, //QE just for reflexivity, no problem
      chase(1) &
        DebuggingTactics.print("Before normalizing") &  prop &
        OnAll(prepareArith2) &
        printGoal & linearElim(linear) & genWitnessTac(ineq,witness))

    val res = proveBy(i, tactic)
    res shouldBe 'proved
  }

  "InvariantArith" should "prove a nasty arithmetic problem from L4Q2" taggedAs SlowTest in withMathematica { qeTool =>
    val antes = IndexedSeq("minr>0","buffer>0","trackr>=minr","(tx-x)^2+(ty-y)^2=trackr^2","(tx-rx)^2+(ty-ry)^2>=(buffer+trackr)^2").map(_.asFormula)
    val succ = IndexedSeq("(x-rx)^2+(y-ry)^2>=buffer^2".asFormula)

    val linear = List((5,"trackr","wit_^2 + minr","1"),(2,"buffer","wit__1^2","1"),(2,"minr","wit__2^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm))
    val witness = List(("4","wit_*wit__0*wit__1"),("4","wit__2*wit__0*wit__1"),("2","wit__0*wit__3"),("4","-ty*x + tx*y + ty*rx - y*rx - tx*ry + x*ry"),("4","wit_*wit__1*wit__3"),("4","wit__2*wit__1*wit__3"),("4","wit_^2*wit__0 + wit__2^2*wit__0"),("1","wit__3^2")).map( s => (s._1.asTerm,s._2.asTerm))
    // The first inequation is used here
    val ineq = List(0)
    //The default basis isn't anywhere near to a groebner basis...
    val insts = Some(List((0,"-2*tx*x - x^2 - 2*ty*y - y^2 + 2*tx*rx + 4*x*rx - 3*rx^2 + 2*ty*ry + 4*y*ry - 3*ry^2 - 2*wit_^2*wit__1^2 - 2*wit__2^2*wit__1^2 - wit__1^4 - wit__3^2"), (1," 4*wit_^4 + 8*wit_^2*wit__2^2 + 4*wit__2^4 + 4*wit_^2*wit__1^2 + 4*wit__2^2*wit__1^2 + 2*wit__3^2"), (2," 2*tx*x - 3*x^2 + 2*ty*y - 3*y^2 - 2*tx*rx + 4*x*rx - rx^2 - 2*ty*ry + 4*y*ry - ry^2 + 2*wit_^2*wit__1^2 + 2*wit__2^2*wit__1^2 + wit__1^4 + wit__3^2")).map (s => (s._1,s._2.asTerm)))

    val tactic = prepareArith2 & DebuggingTactics.print("after") &
      printGoal & linearElim(linear) & genWitnessTac(ineq,witness,insts)

    val res = proveBy(Sequent(antes,succ), tactic)

    res shouldBe 'proved
  }

  "ACASX Arith" should "prove ACAS X 175 fml (division rewritten by hand)" taggedAs IgnoreInBuildTest in withMathematica { qeTool =>
    val i = "(w_0*dhf_0>=0&r_0 < -rp_0&(0>=w_0*(dhf_0-dhd_0)&max_1=0|0 < w_0*(dhf_0-dhd_0)&max_1=w_0*(dhf_0-dhd_0))&(0<=w_0*dhd_0&min_1=0|0>w_0*dhd_0&min_1=w_0*dhd_0)&(r_0-ro_0>=0&abs_1=r_0-ro_0|r_0-ro_0 < 0&abs_1=-(r_0-ro_0))&(w_0=-1|w_0=1)&hp_0>0&rp_0>=0&rv_0>=0&a_0>0&(0<=t_0&a_0 * t_0 < max_1&ro_0=rv_0*t_0&ho_0=w_0*a_0/2*t_0^2+dhd_0*t_0|a_0*t_0>=max_1&ro_0=rv_0*t_0&ho_0*2*a_0=dhf_0*t_0*2*a_0-w_0*max_1^2)->abs_1>rp_0|w_0*h_0 < w_0*ho_0-hp_0)".asFormula
    val lwits = List(Some((List(0),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(5,"a_0","wit__1^2","1"),(9,"dhd_0","-wit__6^2 + dhf_0","-1"),(0,"ho_0","-1/2*wit__1^2*wit__9^4 - (wit__6^2 - dhf_0)*wit__9^2","1"),(0,"abs_1","-wit__10^2*wit__9^2 + r_0","1"),(4,"r_0","-wit__3^2 - rp_0","1"),(0,"min_1","0","1"),(0,"max_1","0","1"),(1,"hp_0","wit__2^2","1"),(1,"rp_0","-1/2*wit__10^2*wit__9^2 - 1/2*wit__3^2 + 1/2*wit__4^2","2"),(1,"h_0","-1/2*wit__1^2*wit__9^4 - (wit__6^2 - dhf_0)*wit__9^2 + wit__2^2 - wit__5^2","-1"),(1,"dhf_0","wit__6^2 - wit__7^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0*wit__1*wit__9")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"-wit__1^2*wit__9^2"),(1,"0"),(2,"0"),(3,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0,2),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(4,"a_0","wit__0^2","1"),(8,"dhf_0","wit__5^2 + dhd_0","1"),(7,"ho_0","wit__4^2 + h_0 - hp_0","1"),(3,"max_1","0","1"),(1,"abs_1","-t_0*wit__9^2 + r_0","1"),(3,"r_0","-wit__2^2 - rp_0","1"),(1,"min_1","0","1"),(1,"hp_0","wit__1^2","1"),(1,"rp_0","-1/2*t_0*wit__9^2 - 1/2*wit__2^2 + 1/2*wit__3^2","2"),(1,"dhd_0","-wit__6^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0^2*wit__2*wit__7"),("1","wit__0*wit__2*wit__9*wit__8"),("1","wit__0^2*wit__2*wit__10")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__0^4*wit__2^2"),(2,"wit__0^2*wit__2^2*wit__9^2"),(3,"wit__0^4*wit__2^2"),(4,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(5,"a_0","wit__1^2","1"),(9,"dhd_0","wit__6^2 + dhf_0","1"),(0,"ho_0","1/2*wit__1^2*wit__9^4 + (wit__6^2 + dhf_0)*wit__9^2","1"),(0,"abs_1","-wit__10^2*wit__9^2 + r_0","1"),(4,"r_0","-wit__3^2 - rp_0","1"),(0,"min_1","0","1"),(0,"max_1","0","1"),(1,"hp_0","wit__2^2","1"),(1,"rp_0","-1/2*wit__10^2*wit__9^2 - 1/2*wit__3^2 + 1/2*wit__4^2","2"),(1,"h_0","1/2*wit__1^2*wit__9^4 + (wit__6^2 + dhf_0)*wit__9^2 - wit__2^2 + wit__5^2","1"),(1,"dhf_0","-wit__6^2 + wit__7^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0*wit__1*wit__9")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"-wit__1^2*wit__9^2"),(1,"0"),(2,"0"),(3,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0,2),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(4,"a_0","wit__0^2","1"),(8,"dhf_0","-wit__5^2 + dhd_0","-1"),(7,"ho_0","-wit__4^2 + h_0 + hp_0","-1"),(3,"max_1","0","1"),(1,"abs_1","-t_0*wit__9^2 + r_0","1"),(3,"r_0","-wit__2^2 - rp_0","1"),(1,"min_1","0","1"),(1,"hp_0","wit__1^2","1"),(1,"rp_0","-1/2*t_0*wit__9^2 - 1/2*wit__2^2 + 1/2*wit__3^2","2"),(1,"dhd_0","wit__6^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0^2*wit__2*wit__7"),("1","wit__0*wit__2*wit__9*wit__8"),("1","wit__0^2*wit__2*wit__10")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__0^4*wit__2^2"),(2,"wit__0^2*wit__2^2*wit__9^2"),(3,"wit__0^4*wit__2^2"),(4,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(1),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(6,"a_0","wit__2^2","1"),(10,"dhd_0","-wit__7^2 + dhf_0","-1"),(0,"ho_0","-1/2*wit__2^2*wit__9^4 - (wit__7^2 - dhf_0)*wit__9^2","1"),(0,"abs_1","wit__10^2*wit__9^2 - r_0","1"),(2,"r_0","wit__10^2*wit__9^2 - wit__0^2","1"),(0,"min_1","0","1"),(0,"max_1","0","1"),(1,"hp_0","wit__3^2","1"),(1,"rp_0","-wit__10^2*wit__9^2 + wit__0^2 - wit__4^2","1"),(2,"h_0","-1/2*wit__2^2*wit__9^4 - (wit__7^2 - dhf_0)*wit__9^2 + wit__3^2 - wit__6^2","-1"),(2,"dhf_0","wit__7^2 - wit__8^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__1*wit__2*wit__9")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"-wit__2^2*wit__9^2"),(1,"0"),(2,"0"),(3,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(1,3),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(5,"a_0","wit__1^2","1"),(9,"dhf_0","wit__6^2 + dhd_0","1"),(8,"ho_0","wit__5^2 + h_0 - hp_0","1"),(3,"max_1","0","1"),(1,"abs_1","t_0*wit__9^2 - r_0","1"),(2,"r_0","t_0*wit__9^2 - wit__0^2","1"),(1,"min_1","0","1"),(1,"hp_0","wit__2^2","1"),(1,"rp_0","-t_0*wit__9^2 + wit__0^2 - wit__3^2","1"),(2,"dhd_0","-wit__7^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__1*wit__9*wit__8*wit__0"),("1","wit__1*wit__3*wit__9*wit__8"),("1","wit__1^2*wit__3*wit__10"),("1","wit__1^2*wit__4*wit__0")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__1^4*wit__0^2"),(2,"wit__1^2*wit__3^2*wit__9^2 + wit__1^2*wit__9^2*wit__0^2"),(3,"wit__1^4*wit__3^2"),(4,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(1),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(6,"a_0","wit__2^2","1"),(10,"dhd_0","wit__7^2 + dhf_0","1"),(0,"ho_0","1/2*wit__2^2*wit__9^4 + (wit__7^2 + dhf_0)*wit__9^2","1"),(0,"abs_1","wit__10^2*wit__9^2 - r_0","1"),(2,"r_0","wit__10^2*wit__9^2 - wit__0^2","1"),(0,"min_1","0","1"),(0,"max_1","0","1"),(1,"hp_0","wit__3^2","1"),(1,"rp_0","-wit__10^2*wit__9^2 + wit__0^2 - wit__4^2","1"),(2,"h_0","1/2*wit__2^2*wit__9^4 + (wit__7^2 + dhf_0)*wit__9^2 - wit__3^2 + wit__6^2","1"),(2,"dhf_0","-wit__7^2 + wit__8^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__1*wit__2*wit__9")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"-wit__2^2*wit__9^2"),(1,"0"),(2,"0"),(3,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(1,3),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(5,"a_0","wit__1^2","1"),(9,"dhf_0","-wit__6^2 + dhd_0","-1"),(8,"ho_0","-wit__5^2 + h_0 + hp_0","-1"),(3,"max_1","0","1"),(1,"abs_1","t_0*wit__9^2 - r_0","1"),(2,"r_0","t_0*wit__9^2 - wit__0^2","1"),(1,"min_1","0","1"),(1,"hp_0","wit__2^2","1"),(1,"rp_0","-t_0*wit__9^2 + wit__0^2 - wit__3^2","1"),(2,"dhd_0","wit__7^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__1*wit__9*wit__8*wit__0"),("1","wit__1*wit__3*wit__9*wit__8"),("1","wit__1^2*wit__3*wit__10"),("1","wit__1^2*wit__4*wit__0")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__1^4*wit__0^2"),(2,"wit__1^2*wit__3^2*wit__9^2 + wit__1^2*wit__9^2*wit__0^2"),(3,"wit__1^4*wit__3^2"),(4,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(6,"a_0","wit__2^2","1"),(2,"dhd_0","-min_1","1"),(0,"ho_0","-1/2*wit__2^2*wit__9^4 - min_1*wit__9^2","1"),(0,"abs_1","-wit__10^2*wit__9^2 + r_0","1"),(4,"r_0","-wit__4^2 - rp_0","1"),(1,"min_1","-wit__0^2","-1"),(0,"max_1","0","1"),(1,"hp_0","wit__3^2","1"),(1,"rp_0","-1/2*wit__10^2*wit__9^2 - 1/2*wit__4^2 + 1/2*wit__5^2","2"),(1,"h_0","-1/2*wit__2^2*wit__9^4 + wit__0^2*wit__9^2 + wit__3^2 - wit__6^2","-1"),(1,"dhf_0","wit__0^2 + wit__7^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit_*wit__0"),("1","wit__0*wit__7")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"wit__0^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(5,"a_0","wit__1^2","1"),(9,"dhf_0","wit__6^2 + dhd_0","1"),(8,"ho_0","wit__5^2 + h_0 - hp_0","1"),(3,"max_1","0","1"),(1,"abs_1","-t_0*wit__9^2 + r_0","1"),(4,"r_0","-wit__3^2 - rp_0","1"),(1,"dhd_0","-min_1","1"),(1,"min_1","-wit__0^2","-1"),(1,"hp_0","wit__2^2","1"),(1,"rp_0","-1/2*t_0*wit__9^2 - 1/2*wit__3^2 + 1/2*wit__4^2","2")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0*wit__6"),("1","wit__0*wit_")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"0"),(4,"wit__0^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(6,"a_0","wit__2^2","1"),(2,"dhd_0","min_1","-1"),(0,"ho_0","1/2*wit__2^2*wit__9^4 + min_1*wit__9^2","1"),(0,"abs_1","-wit__10^2*wit__9^2 + r_0","1"),(4,"r_0","-wit__4^2 - rp_0","1"),(1,"min_1","-wit__0^2","-1"),(0,"max_1","0","1"),(1,"hp_0","wit__3^2","1"),(1,"rp_0","-1/2*wit__10^2*wit__9^2 - 1/2*wit__4^2 + 1/2*wit__5^2","2"),(1,"h_0","1/2*wit__2^2*wit__9^4 - wit__0^2*wit__9^2 - wit__3^2 + wit__6^2","1"),(1,"dhf_0","-wit__0^2 - wit__7^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit_*wit__0"),("1","wit__0*wit__7")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"wit__0^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(5,"a_0","wit__1^2","1"),(9,"dhf_0","-wit__6^2 + dhd_0","-1"),(8,"ho_0","-wit__5^2 + h_0 + hp_0","-1"),(3,"max_1","0","1"),(1,"abs_1","-t_0*wit__9^2 + r_0","1"),(4,"r_0","-wit__3^2 - rp_0","1"),(1,"dhd_0","min_1","-1"),(1,"min_1","-wit__0^2","-1"),(1,"hp_0","wit__2^2","1"),(1,"rp_0","-1/2*t_0*wit__9^2 - 1/2*wit__3^2 + 1/2*wit__4^2","2")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0*wit__6"),("1","wit__0*wit_")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"0"),(4,"wit__0^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(7,"a_0","wit__3^2","1"),(2,"dhd_0","-min_1","1"),(0,"ho_0","-1/2*wit__3^2*wit__9^4 - min_1*wit__9^2","1"),(0,"abs_1","wit__10^2*wit__9^2 - r_0","1"),(2,"r_0","wit__10^2*wit__9^2 - wit__1^2","1"),(1,"min_1","-wit__0^2","-1"),(0,"max_1","0","1"),(1,"hp_0","wit__4^2","1"),(1,"rp_0","-wit__10^2*wit__9^2 + wit__1^2 - wit__5^2","1"),(2,"h_0","-1/2*wit__3^2*wit__9^4 + wit__0^2*wit__9^2 + wit__4^2 - wit__7^2","-1"),(2,"dhf_0","wit__0^2 + wit__8^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit_*wit__0"),("1","wit__0*wit__8")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"wit__0^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(6,"a_0","wit__2^2","1"),(10,"dhf_0","wit__7^2 + dhd_0","1"),(9,"ho_0","wit__6^2 + h_0 - hp_0","1"),(3,"max_1","0","1"),(1,"abs_1","t_0*wit__9^2 - r_0","1"),(3,"r_0","t_0*wit__9^2 - wit__1^2","1"),(1,"dhd_0","-min_1","1"),(1,"min_1","-wit__0^2","-1"),(1,"hp_0","wit__3^2","1"),(1,"rp_0","-t_0*wit__9^2 + wit__1^2 - wit__4^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0*wit__7"),("1","wit__0*wit_")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"0"),(4,"wit__0^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(7,"a_0","wit__3^2","1"),(2,"dhd_0","min_1","-1"),(0,"ho_0","1/2*wit__3^2*wit__9^4 + min_1*wit__9^2","1"),(0,"abs_1","wit__10^2*wit__9^2 - r_0","1"),(2,"r_0","wit__10^2*wit__9^2 - wit__1^2","1"),(1,"min_1","-wit__0^2","-1"),(0,"max_1","0","1"),(1,"hp_0","wit__4^2","1"),(1,"rp_0","-wit__10^2*wit__9^2 + wit__1^2 - wit__5^2","1"),(2,"h_0","1/2*wit__3^2*wit__9^4 - wit__0^2*wit__9^2 - wit__4^2 + wit__7^2","1"),(2,"dhf_0","-wit__0^2 - wit__8^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit_*wit__0"),("1","wit__0*wit__8")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"wit__0^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(6,"a_0","wit__2^2","1"),(10,"dhf_0","-wit__7^2 + dhd_0","-1"),(9,"ho_0","-wit__6^2 + h_0 + hp_0","-1"),(3,"max_1","0","1"),(1,"abs_1","t_0*wit__9^2 - r_0","1"),(3,"r_0","t_0*wit__9^2 - wit__1^2","1"),(1,"dhd_0","min_1","-1"),(1,"min_1","-wit__0^2","-1"),(1,"hp_0","wit__3^2","1"),(1,"rp_0","-t_0*wit__9^2 + wit__1^2 - wit__4^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0*wit__7"),("1","wit__0*wit_")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"0"),(4,"wit__0^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(4),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(6,"a_0","wit__2^2","1"),(3,"dhd_0","dhf_0 + max_1","-1"),(0,"ho_0","-1/2*wit__2^2*wit__9^4 + (dhf_0 + max_1)*wit__9^2","1"),(0,"abs_1","-wit__10^2*wit__9^2 + r_0","1"),(4,"r_0","-wit__4^2 - rp_0","1"),(0,"min_1","0","1"),(5,"dhf_0","-wit__7^2 - max_1","1"),(0,"max_1","wit__0^2","-1"),(1,"hp_0","wit__3^2","1"),(1,"rp_0","-1/2*wit__10^2*wit__9^2 - 1/2*wit__4^2 + 1/2*wit__5^2","2"),(1,"h_0","-1/2*wit__2^2*wit__9^4 - wit__7^2*wit__9^2 + wit__3^2 - wit__6^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__9*wit__10*wit__4"),("1","wit__4*wit__8"),("1","wit__4*wit__11")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__4^2"),(2,"wit__4^2"),(3,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(1,3),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(5,"a_0","wit__1^2","1"),(3,"dhf_0","dhd_0 - max_1","1"),(7,"ho_0","wit__5^2 + h_0 - hp_0","1"),(3,"max_1","wit__0^2","-1"),(1,"abs_1","-t_0*wit__9^2 + r_0","1"),(3,"r_0","-wit__3^2 - rp_0","1"),(1,"min_1","0","1"),(3,"dhd_0","-wit__6^2","1"),(1,"hp_0","wit__2^2","1"),(1,"rp_0","-1/2*t_0*wit__9^2 - 1/2*wit__3^2 + 1/2*wit__4^2","2")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__1*wit__6*wit__7*wit__9"),("1","wit__6*wit__9^2*wit__8"),("1","wit__1*wit__6*wit__9*wit__10"),("1","wit__1^2*wit__3*wit__7"),("1","wit__1*wit__3*wit__9*wit__8"),("1","wit__1^2*wit__3*wit__10"),("1","wit__1*wit__3*wit__9*wit_"),("1","wit__0*wit__6*wit__9^2")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__1^4*wit__3^2 + wit__1^2*wit__6^2*wit__9^2"),(2,"wit__1^2*wit__3^2*wit__9^2 + wit__6^2*wit__9^4"),(3,"wit__1^4*wit__3^2 + wit__1^2*wit__6^2*wit__9^2"),(4,"wit__1^2*wit__3^2*wit__9^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(4),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(6,"a_0","wit__2^2","1"),(3,"dhd_0","dhf_0 - max_1","1"),(0,"ho_0","1/2*wit__2^2*wit__9^4 + (dhf_0 - max_1)*wit__9^2","1"),(0,"abs_1","-wit__10^2*wit__9^2 + r_0","1"),(4,"r_0","-wit__4^2 - rp_0","1"),(0,"min_1","0","1"),(5,"dhf_0","wit__7^2 + max_1","-1"),(0,"max_1","wit__0^2","-1"),(1,"hp_0","wit__3^2","1"),(1,"rp_0","-1/2*wit__10^2*wit__9^2 - 1/2*wit__4^2 + 1/2*wit__5^2","2"),(1,"h_0","1/2*wit__2^2*wit__9^4 + wit__7^2*wit__9^2 - wit__3^2 + wit__6^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__9*wit__10*wit__4"),("1","wit__4*wit__8"),("1","wit__4*wit__11")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__4^2"),(2,"wit__4^2"),(3,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(1,3),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(5,"a_0","wit__1^2","1"),(3,"dhf_0","dhd_0 + max_1","-1"),(7,"ho_0","-wit__5^2 + h_0 + hp_0","-1"),(3,"max_1","wit__0^2","-1"),(1,"abs_1","-t_0*wit__9^2 + r_0","1"),(3,"r_0","-wit__3^2 - rp_0","1"),(1,"min_1","0","1"),(3,"dhd_0","wit__6^2","-1"),(1,"hp_0","wit__2^2","1"),(1,"rp_0","-1/2*t_0*wit__9^2 - 1/2*wit__3^2 + 1/2*wit__4^2","2")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__1*wit__6*wit__7*wit__9"),("1","wit__6*wit__9^2*wit__8"),("1","wit__1*wit__6*wit__9*wit__10"),("1","wit__1^2*wit__3*wit__7"),("1","wit__1*wit__3*wit__9*wit__8"),("1","wit__1^2*wit__3*wit__10"),("1","wit__1*wit__3*wit__9*wit_"),("1","wit__0*wit__6*wit__9^2")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__1^4*wit__3^2 + wit__1^2*wit__6^2*wit__9^2"),(2,"wit__1^2*wit__3^2*wit__9^2 + wit__6^2*wit__9^4"),(3,"wit__1^4*wit__3^2 + wit__1^2*wit__6^2*wit__9^2"),(4,"wit__1^2*wit__3^2*wit__9^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(5),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(7,"a_0","wit__3^2","1"),(3,"dhd_0","dhf_0 + max_1","-1"),(0,"ho_0","-1/2*wit__3^2*wit__9^4 + (dhf_0 + max_1)*wit__9^2","1"),(0,"abs_1","wit__10^2*wit__9^2 - r_0","1"),(2,"r_0","wit__10^2*wit__9^2 - wit__1^2","1"),(0,"min_1","0","1"),(6,"dhf_0","-wit__8^2 - max_1","1"),(0,"max_1","wit__0^2","-1"),(1,"hp_0","wit__4^2","1"),(1,"rp_0","-wit__10^2*wit__9^2 + wit__1^2 - wit__5^2","1"),(2,"h_0","-1/2*wit__3^2*wit__9^4 - wit__8^2*wit__9^2 + wit__4^2 - wit__7^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__9*wit__10*wit__5"),("1","wit__5*wit__6")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__5^2"),(2,"0"),(3,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(2,4),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(6,"a_0","wit__2^2","1"),(3,"dhf_0","dhd_0 - max_1","1"),(8,"ho_0","wit__6^2 + h_0 - hp_0","1"),(3,"max_1","wit__0^2","-1"),(1,"abs_1","t_0*wit__9^2 - r_0","1"),(2,"r_0","t_0*wit__9^2 - wit__1^2","1"),(1,"min_1","0","1"),(4,"dhd_0","-wit__7^2","1"),(1,"hp_0","wit__3^2","1"),(1,"rp_0","-t_0*wit__9^2 + wit__1^2 - wit__4^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__7*wit__9^2*wit__8"),("1","wit__2*wit__9*wit__8*wit__1"),("1","wit__2*wit__4*wit__9*wit__8"),("1","wit__2^2*wit__4*wit__10"),("1","wit__2*wit__4*wit__9*wit_"),("1","wit__0*wit__7*wit__9^2"),("1","wit__0*wit__2*wit__9*wit__1"),("1","wit__2*wit__7*wit__5*wit__9"),("1","wit__2^2*wit__5*wit__1")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__2^2*wit__7^2*wit__9^2 + wit__2^4*wit__1^2"),(2,"wit__2^2*wit__4^2*wit__9^2 + wit__7^2*wit__9^4 + wit__2^2*wit__9^2*wit__1^2"),(3,"wit__2^4*wit__4^2"),(4,"wit__2^2*wit__4^2*wit__9^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(5),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(7,"a_0","wit__3^2","1"),(3,"dhd_0","dhf_0 - max_1","1"),(0,"ho_0","1/2*wit__3^2*wit__9^4 + (dhf_0 - max_1)*wit__9^2","1"),(0,"abs_1","wit__10^2*wit__9^2 - r_0","1"),(2,"r_0","wit__10^2*wit__9^2 - wit__1^2","1"),(0,"min_1","0","1"),(6,"dhf_0","wit__8^2 + max_1","-1"),(0,"max_1","wit__0^2","-1"),(1,"hp_0","wit__4^2","1"),(1,"rp_0","-wit__10^2*wit__9^2 + wit__1^2 - wit__5^2","1"),(2,"h_0","1/2*wit__3^2*wit__9^4 + wit__8^2*wit__9^2 - wit__4^2 + wit__7^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__9*wit__10*wit__5"),("1","wit__5*wit__6")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__5^2"),(2,"0"),(3,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(2,4),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(6,"a_0","wit__2^2","1"),(3,"dhf_0","dhd_0 + max_1","-1"),(8,"ho_0","-wit__6^2 + h_0 + hp_0","-1"),(3,"max_1","wit__0^2","-1"),(1,"abs_1","t_0*wit__9^2 - r_0","1"),(2,"r_0","t_0*wit__9^2 - wit__1^2","1"),(1,"min_1","0","1"),(4,"dhd_0","wit__7^2","-1"),(1,"hp_0","wit__3^2","1"),(1,"rp_0","-t_0*wit__9^2 + wit__1^2 - wit__4^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__7*wit__9^2*wit__8"),("1","wit__2*wit__9*wit__8*wit__1"),("1","wit__2*wit__4*wit__9*wit__8"),("1","wit__2^2*wit__4*wit__10"),("1","wit__2*wit__4*wit__9*wit_"),("1","wit__0*wit__7*wit__9^2"),("1","wit__0*wit__2*wit__9*wit__1"),("1","wit__2*wit__7*wit__5*wit__9"),("1","wit__2^2*wit__5*wit__1")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__2^2*wit__7^2*wit__9^2 + wit__2^4*wit__1^2"),(2,"wit__2^2*wit__4^2*wit__9^2 + wit__7^2*wit__9^4 + wit__2^2*wit__9^2*wit__1^2"),(3,"wit__2^4*wit__4^2"),(4,"wit__2^2*wit__4^2*wit__9^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(5),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(7,"a_0","wit__3^2","1"),(2,"dhd_0","-min_1","1"),(0,"ho_0","-1/2*wit__3^2*wit__9^4 - min_1*wit__9^2","1"),(0,"abs_1","-wit__10^2*wit__9^2 + r_0","1"),(5,"r_0","-wit__5^2 - rp_0","1"),(0,"min_1","-dhf_0 - max_1","1"),(1,"dhf_0","wit__1^2 - max_1","1"),(0,"max_1","wit__0^2","-1"),(1,"hp_0","wit__4^2","1"),(1,"rp_0","-1/2*wit__10^2*wit__9^2 - 1/2*wit__5^2 + 1/2*wit__6^2","2"),(1,"h_0","-1/2*wit__3^2*wit__9^4 + wit__1^2*wit__9^2 + wit__4^2 - wit__7^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__9*wit__10*wit__5"),("1","wit__5*wit__8"),("1","wit__5*wit__11")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__5^2"),(2,"wit__5^2"),(3,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(2,4),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(6,"a_0","wit__2^2","1"),(3,"dhf_0","dhd_0 - max_1","1"),(8,"ho_0","wit__6^2 + h_0 - hp_0","1"),(3,"max_1","wit__0^2","-1"),(1,"abs_1","-t_0*wit__9^2 + r_0","1"),(4,"r_0","-wit__4^2 - rp_0","1"),(1,"dhd_0","-min_1","1"),(1,"min_1","-wit__1^2","-1"),(1,"hp_0","wit__3^2","1"),(1,"rp_0","-1/2*t_0*wit__9^2 - 1/2*wit__4^2 + 1/2*wit__5^2","2")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__2^2*wit__4*wit__7"),("1","wit__1*wit__2*wit__4*wit__9"),("1","wit__2*wit__4*wit__9*wit__8"),("1","wit__2^2*wit__4*wit__10"),("1","wit__2*wit__4*wit__9*wit_")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__2^4*wit__4^2"),(2,"wit__2^2*wit__4^2*wit__9^2"),(3,"wit__2^4*wit__4^2"),(4,"wit__2^2*wit__4^2*wit__9^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(5),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(7,"a_0","wit__3^2","1"),(2,"dhd_0","min_1","-1"),(0,"ho_0","1/2*wit__3^2*wit__9^4 + min_1*wit__9^2","1"),(0,"abs_1","-wit__10^2*wit__9^2 + r_0","1"),(5,"r_0","-wit__5^2 - rp_0","1"),(0,"min_1","dhf_0 - max_1","1"),(1,"dhf_0","-wit__1^2 + max_1","-1"),(0,"max_1","wit__0^2","-1"),(1,"hp_0","wit__4^2","1"),(1,"rp_0","-1/2*wit__10^2*wit__9^2 - 1/2*wit__5^2 + 1/2*wit__6^2","2"),(1,"h_0","1/2*wit__3^2*wit__9^4 - wit__1^2*wit__9^2 - wit__4^2 + wit__7^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__9*wit__10*wit__5"),("1","wit__5*wit__8"),("1","wit__5*wit__11")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__5^2"),(2,"wit__5^2"),(3,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(2,4),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(6,"a_0","wit__2^2","1"),(3,"dhf_0","dhd_0 + max_1","-1"),(8,"ho_0","-wit__6^2 + h_0 + hp_0","-1"),(3,"max_1","wit__0^2","-1"),(1,"abs_1","-t_0*wit__9^2 + r_0","1"),(4,"r_0","-wit__4^2 - rp_0","1"),(1,"dhd_0","min_1","-1"),(1,"min_1","-wit__1^2","-1"),(1,"hp_0","wit__3^2","1"),(1,"rp_0","-1/2*t_0*wit__9^2 - 1/2*wit__4^2 + 1/2*wit__5^2","2")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__2^2*wit__4*wit__7"),("1","wit__1*wit__2*wit__4*wit__9"),("1","wit__2*wit__4*wit__9*wit__8"),("1","wit__2^2*wit__4*wit__10"),("1","wit__2*wit__4*wit__9*wit_")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__2^4*wit__4^2"),(2,"wit__2^2*wit__4^2*wit__9^2"),(3,"wit__2^4*wit__4^2"),(4,"wit__2^2*wit__4^2*wit__9^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(6),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(8,"a_0","wit__4^2","1"),(2,"dhd_0","-min_1","1"),(0,"ho_0","-1/2*wit__4^2*wit__9^4 - min_1*wit__9^2","1"),(0,"abs_1","wit__10^2*wit__9^2 - r_0","1"),(3,"r_0","wit__10^2*wit__9^2 - wit__2^2","1"),(0,"min_1","-dhf_0 - max_1","1"),(1,"dhf_0","wit__1^2 - max_1","1"),(0,"max_1","wit__0^2","-1"),(1,"hp_0","wit__5^2","1"),(1,"rp_0","-wit__10^2*wit__9^2 + wit__2^2 - wit__6^2","1"),(2,"h_0","-1/2*wit__4^2*wit__9^4 + wit__1^2*wit__9^2 + wit__5^2 - wit__8^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__9*wit__10*wit__6"),("1","wit__6*wit__7")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__6^2"),(2,"0"),(3,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(3,5),List((0,"w_0","-1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(7,"a_0","wit__3^2","1"),(3,"dhf_0","dhd_0 - max_1","1"),(9,"ho_0","wit__7^2 + h_0 - hp_0","1"),(3,"max_1","wit__0^2","-1"),(1,"abs_1","t_0*wit__9^2 - r_0","1"),(3,"r_0","t_0*wit__9^2 - wit__2^2","1"),(1,"dhd_0","-min_1","1"),(1,"min_1","-wit__1^2","-1"),(1,"hp_0","wit__4^2","1"),(1,"rp_0","-t_0*wit__9^2 + wit__2^2 - wit__5^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__1*wit__3*wit__9*wit__2"),("1","wit__3*wit__9*wit__8*wit__2"),("1","wit__3*wit__9*wit__2*wit_"),("1","wit__1*wit__3*wit__5*wit__9"),("1","wit__3*wit__5*wit__9*wit__8"),("1","wit__3^2*wit__5*wit__10"),("1","wit__3*wit__5*wit__9*wit_"),("1","wit__3^2*wit__6*wit__2")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__3^4*wit__2^2"),(2,"wit__3^2*wit__5^2*wit__9^2 + wit__3^2*wit__9^2*wit__2^2"),(3,"wit__3^4*wit__5^2"),(4,"wit__3^2*wit__5^2*wit__9^2 + wit__3^2*wit__9^2*wit__2^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(6),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(14,"rv_0","wit__10^2","1"),(13,"t_0","wit__9^2","-1"),(8,"a_0","wit__4^2","1"),(2,"dhd_0","min_1","-1"),(0,"ho_0","1/2*wit__4^2*wit__9^4 + min_1*wit__9^2","1"),(0,"abs_1","wit__10^2*wit__9^2 - r_0","1"),(3,"r_0","wit__10^2*wit__9^2 - wit__2^2","1"),(0,"min_1","dhf_0 - max_1","1"),(1,"dhf_0","-wit__1^2 + max_1","-1"),(0,"max_1","wit__0^2","-1"),(1,"hp_0","wit__5^2","1"),(1,"rp_0","-wit__10^2*wit__9^2 + wit__2^2 - wit__6^2","1"),(2,"h_0","1/2*wit__4^2*wit__9^4 - wit__1^2*wit__9^2 - wit__5^2 + wit__8^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__9*wit__10*wit__6"),("1","wit__6*wit__7")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__6^2"),(2,"0"),(3,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(3,5),List((0,"w_0","1","1"),(0,"ro_0","rv_0*t_0","1"),(13,"rv_0","wit__9^2","1"),(7,"a_0","wit__3^2","1"),(3,"dhf_0","dhd_0 + max_1","-1"),(9,"ho_0","-wit__7^2 + h_0 + hp_0","-1"),(3,"max_1","wit__0^2","-1"),(1,"abs_1","t_0*wit__9^2 - r_0","1"),(3,"r_0","t_0*wit__9^2 - wit__2^2","1"),(1,"dhd_0","min_1","-1"),(1,"min_1","-wit__1^2","-1"),(1,"hp_0","wit__4^2","1"),(1,"rp_0","-t_0*wit__9^2 + wit__2^2 - wit__5^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__1*wit__3*wit__9*wit__2"),("1","wit__3*wit__9*wit__8*wit__2"),("1","wit__3*wit__9*wit__2*wit_"),("1","wit__1*wit__3*wit__5*wit__9"),("1","wit__3*wit__5*wit__9*wit__8"),("1","wit__3^2*wit__5*wit__10"),("1","wit__3*wit__5*wit__9*wit_"),("1","wit__3^2*wit__6*wit__2")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"wit__3^4*wit__2^2"),(2,"wit__3^2*wit__5^2*wit__9^2 + wit__3^2*wit__9^2*wit__2^2"),(3,"wit__3^4*wit__5^2"),(4,"wit__3^2*wit__5^2*wit__9^2 + wit__3^2*wit__9^2*wit__2^2")).map (s => (s._1,s._2.asTerm))))))

    val tactic = implyR(1) & DebuggingTactics.print("Before normalizing") & prop &
      DebuggingTactics.print("Before split") &
      OnAll(prepareArith2) &
      printGoal

    val res = proveBy(i, tactic)

    val result = lwits.zipWithIndex.map(
      prind =>
        prind._1 match
        {
          case Some(w) =>
            {
              val _ = println("SOLVING",prind._2)
              proveBy(res.subgoals(prind._2), linearElim(w._2) & genWitnessTac(w._1,w._3,w._4) & done)
            }
        }
      )

    print(result)
  }

  "ACASX Arith" should "prove ACAS X 5614 fml (division rewritten by hand)" taggedAs IgnoreInBuildTest in withMathematica { qeTool =>
    val i = "(w_0*dhf_0>=0&(0>=w_0*(dhf_0-dhd_0)&max_1=0|0 < w_0*(dhf_0-dhd_0)&max_1=w_0*(dhf_0-dhd_0))&(0<=w_0*dhd_0&min_1=0|0>w_0*dhd_0&min_1=w_0*dhd_0)&(r_0-rv_0*t_0>=0&abs_1=r_0-rv_0*t_0|r_0-rv_0*t_0 < 0&abs_1=-(r_0-rv_0*t_0))&w_0*rv_0^2*h_0 < a_0/2*(r_0-rp_0)^2+w_0*rv_0*dhd_0*(r_0-rp_0)-rv_0^2*hp_0&rv_0*t_0<=(r_0-rp_0)&(w_0=-1|w_0=1)&a_0*rp_0-rv_0*min_1 < a_0*r_0&a_0*r_0<=a_0*rp_0+rv_0*max_1&a_0*t_0>=max_1&hp_0>0&rp_0>=0&rv_0>=0&a_0>0->abs_1>rp_0|w_0*h_0*(2*a_0) < w_0*(dhf_0*t_0*(2*a_0)-w_0*max_1^2)-(2*a_0)*hp_0)".asFormula
    val lwits = List[Option[(List[Int], List[(Int, Term, Term, Term)], List[(Term, Term)], Option[List[(Int, Term)]])]](Some((List(2),List((0,"w_0","-1","1"),(0,"abs_1","-rv_0*t_0 + r_0","1"),(6,"r_0","rv_0*t_0 - wit__4^2 + rp_0","-1"),(10,"rv_0","wit__9^2","1"),(0,"min_1","0","1"),(0,"max_1","0","1"),(0,"a_0","wit__0^2","1"),(0,"hp_0","wit__1^2","1"),(5,"rp_0","wit__4^2 + wit__8^2","1"),(3,"dhd_0","-wit__6^2 + dhf_0","-1"),(3,"dhf_0","wit__6^2 - wit__7^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__2*wit__12")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"-wit__12^2"),(1,"0"),(2,"0"),(3,"0"),(4,"-t_0*wit__0^2*wit__9^4 - (wit__0^2*wit__13^2 + wit__12^2)*wit__9^2"),(5,"-wit__0^2*wit__13^2 - wit__11^2*wit__9^2"),(6,"-wit__0^4*wit__4^2 + wit__0^2*wit__12^2 + (2*t_0*wit__0^4 - wit__0^2*wit__11^2)*wit__9^2"),(7,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(2),List((0,"w_0","1","1"),(0,"abs_1","-rv_0*t_0 + r_0","1"),(6,"r_0","rv_0*t_0 - wit__4^2 + rp_0","-1"),(10,"rv_0","wit__9^2","1"),(0,"min_1","0","1"),(0,"max_1","0","1"),(0,"a_0","wit__0^2","1"),(0,"hp_0","wit__1^2","1"),(5,"rp_0","wit__4^2 + wit__8^2","1"),(3,"dhd_0","wit__6^2 + dhf_0","1"),(3,"dhf_0","-wit__6^2 + wit__7^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__2*wit__12")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"-wit__12^2"),(1,"0"),(2,"0"),(3,"0"),(4,"-t_0*wit__0^2*wit__9^4 - (wit__0^2*wit__13^2 + wit__12^2)*wit__9^2"),(5,"-wit__0^2*wit__13^2 - wit__11^2*wit__9^2"),(6,"-wit__0^4*wit__4^2 + wit__0^2*wit__12^2 + (2*t_0*wit__0^4 - wit__0^2*wit__11^2)*wit__9^2"),(7,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","-1","1"),(0,"abs_1","rv_0*t_0 - r_0","1"),(2,"r_0","rv_0*t_0 - wit__0^2","1"),(10,"rv_0","wit__9^2","1"),(0,"min_1","0","1"),(0,"max_1","0","1"),(0,"a_0","wit__1^2","1"),(0,"hp_0","wit__2^2","1"),(2,"rp_0","wit__0^2 + wit__5^2","1"),(3,"dhd_0","-wit__7^2 + dhf_0","-1"),(3,"dhf_0","wit__7^2 - wit__8^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0*wit__10"),("1","wit__0*wit__13")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"wit__0^2"),(4,"0"),(5,"0"),(6,"-wit__0^2"),(7,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","1","1"),(0,"abs_1","rv_0*t_0 - r_0","1"),(2,"r_0","rv_0*t_0 - wit__0^2","1"),(10,"rv_0","wit__9^2","1"),(0,"min_1","0","1"),(0,"max_1","0","1"),(0,"a_0","wit__1^2","1"),(0,"hp_0","wit__2^2","1"),(2,"rp_0","wit__0^2 + wit__5^2","1"),(3,"dhd_0","wit__7^2 + dhf_0","1"),(3,"dhf_0","-wit__7^2 + wit__8^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0*wit__10"),("1","wit__0*wit__13")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"wit__0^2"),(4,"0"),(5,"0"),(6,"-wit__0^2"),(7,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","-1","1"),(0,"abs_1","-rv_0*t_0 + r_0","1"),(7,"r_0","rv_0*t_0 - wit__5^2 + rp_0","-1"),(10,"rv_0","wit__9^2","1"),(0,"dhd_0","-min_1","1"),(1,"min_1","-wit__0^2","-1"),(0,"max_1","0","1"),(0,"a_0","wit__1^2","1"),(0,"hp_0","wit__2^2","1"),(4,"rp_0","wit__5^2 + wit__8^2","1"),(3,"dhf_0","wit__0^2 + wit__7^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0*wit__7"),("1","wit__0*wit_")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"0"),(4,"0"),(5,"0"),(6,"0"),(7,"wit__0^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","1","1"),(0,"abs_1","-rv_0*t_0 + r_0","1"),(7,"r_0","rv_0*t_0 - wit__5^2 + rp_0","-1"),(10,"rv_0","wit__9^2","1"),(0,"dhd_0","min_1","-1"),(1,"min_1","-wit__0^2","-1"),(0,"max_1","0","1"),(0,"a_0","wit__1^2","1"),(0,"hp_0","wit__2^2","1"),(4,"rp_0","wit__5^2 + wit__8^2","1"),(3,"dhf_0","-wit__0^2 - wit__7^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0*wit__7"),("1","wit__0*wit_")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"0"),(4,"0"),(5,"0"),(6,"0"),(7,"wit__0^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","-1","1"),(0,"abs_1","rv_0*t_0 - r_0","1"),(3,"r_0","rv_0*t_0 - wit__1^2","1"),(10,"rv_0","wit__9^2","1"),(0,"dhd_0","-min_1","1"),(1,"min_1","-wit__0^2","-1"),(0,"max_1","0","1"),(0,"a_0","wit__2^2","1"),(0,"hp_0","wit__3^2","1"),(2,"rp_0","wit__1^2 + wit__6^2","1"),(3,"dhf_0","wit__0^2 + wit__8^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0*wit__8"),("1","wit__0*wit_")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"0"),(4,"0"),(5,"0"),(6,"0"),(7,"wit__0^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0),List((0,"w_0","1","1"),(0,"abs_1","rv_0*t_0 - r_0","1"),(3,"r_0","rv_0*t_0 - wit__1^2","1"),(10,"rv_0","wit__9^2","1"),(0,"dhd_0","min_1","-1"),(1,"min_1","-wit__0^2","-1"),(0,"max_1","0","1"),(0,"a_0","wit__2^2","1"),(0,"hp_0","wit__3^2","1"),(2,"rp_0","wit__1^2 + wit__6^2","1"),(3,"dhf_0","-wit__0^2 - wit__8^2","-1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__0*wit__8"),("1","wit__0*wit_")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"0"),(4,"0"),(5,"0"),(6,"0"),(7,"wit__0^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0,1,4),List((0,"w_0","-1","1"),(0,"abs_1","-rv_0*t_0 + r_0","1"),(7,"r_0","rv_0*t_0 - wit__5^2 + rp_0","-1"),(10,"rv_0","wit__9^2","1"),(0,"min_1","0","1"),(0,"dhd_0","dhf_0 + max_1","-1"),(6,"dhf_0","-wit__7^2 - max_1","1"),(0,"max_1","wit__0^2","-1"),(0,"a_0","wit__1^2","1"),(0,"hp_0","wit__2^2","1"),(3,"rp_0","wit__5^2 + wit__8^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__1*wit__3^2*wit__4*wit__11*wit_ + wit__1*wit__4*wit__11*wit__12^2*wit_"),("1","wit__1*wit__3*wit__4*wit__0*wit__12*wit_"),("1","wit__1*wit__4*wit__0*wit__12^2*wit_"),("1/2","wit__3^2*wit__11*wit__12^2*wit_ + wit__11*wit__12^4*wit_"),("1/2","wit__3*wit__0*wit__12^3*wit_"),("1/4","wit__0*wit__12^4*wit_"),("1/2","wit__1*wit__3^2*wit__4*wit__6 + wit__1*wit__4*wit__6*wit__12^2"),("1/4","wit__3^2*wit__6*wit__12^2 + wit__6*wit__12^4"),("1/4","wit__7*wit__0*wit__12^4")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"-2*wit_^2*wit__1^2*wit__11^2*wit__12^2*wit__4^2 + 1/4*(4*wit_^4 - 6*wit_^2*wit__0^2 + wit__0^4 - wit__11^4 - 2*h_0*wit__1^2 - (4*wit_^2 + wit__0^2)*wit__11^2)*wit__12^6 + 1/4*(2*wit__1^2*wit__11^4 + (8*wit_^2 - 3*wit__0^2)*wit__1^2*wit__11^2 + 2*(2*wit_^4 - 3*wit_^2*wit__0^2)*wit__1^2)*wit__12^4*wit__13^2 + 1/4*(3*wit__1^4*wit__11^4 + 4*h_0*wit__1^6 + (8*wit_^2 + wit__0^2)*wit__1^4*wit__11^2 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^4)*wit__12^2*wit__13^4 + 1/4*(4*wit_^2*wit__0^2*wit__1^6 + 3*wit__0^2*wit__1^6*wit__11^2)*wit__13^6 - 1/4*(2*wit_^2*wit__0^2 - wit__0^4 + 2*wit_^2*wit__11^2 - 2*wit__1^2*wit__2^2 + 2*h_0*wit__1^2)*wit__3^6 + 1/4*(4*wit_^4*wit__0^6 - 4*wit_^2*wit__0^8 + wit__0^10 + 4*wit_^2*wit__0^2*wit__11^6 + wit__0^2*wit__11^8 + 4*wit__0^2*wit__1^4*wit__2^4 + 4*h_0^2*wit__0^2*wit__1^4 + 2*(2*wit_^4*wit__0^2 + 2*wit_^2*wit__0^4 - wit__0^6 + 2*h_0*wit__0^2*wit__1^2)*wit__11^4 + 4*(2*h_0*wit_^2*wit__0^4 - h_0*wit__0^6)*wit__1^2 + 4*(2*wit_^4*wit__0^4 - wit_^2*wit__0^6 + 2*h_0*wit_^2*wit__0^2*wit__1^2)*wit__11^2 - 4*(2*wit_^2*wit__0^2*wit__1^2*wit__11^2 + wit__0^2*wit__1^2*wit__11^4 + 2*h_0*wit__0^2*wit__1^4 + (2*wit_^2*wit__0^4 - wit__0^6)*wit__1^2)*wit__2^2)*wit__9^6 + 1/4*(6*wit__1^2*wit__12^2*wit__2^2 + (4*wit_^4 - 8*wit_^2*wit__0^2 + 3*wit__0^4 - 6*h_0*wit__1^2 + 2*(2*wit_^2 - wit__0^2)*wit__11^2)*wit__12^2 + 2*((2*wit_^2 - wit__0^2)*wit__1^2*wit__11^2 + (2*wit_^4 - wit_^2*wit__0^2)*wit__1^2)*wit__13^2)*wit__3^4 + (2*wit_^2*wit__1^4*wit__12^2*wit__13^2 + wit_^2*wit__1^2*wit__12^4 - wit_^2*wit__1^2*wit__12^2*wit__3^2)*wit__5^2*wit__7^2 + 1/4*(4*wit__1^4*wit__12^2*wit__2^4 + (4*wit_^4*wit__0^4 - 4*wit_^2*wit__0^6 + wit__0^8 + 4*wit_^2*wit__11^6 + wit__11^8 + 4*h_0^2*wit__1^4 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4 + 2*h_0*wit__1^2)*wit__11^4 + 4*(2*h_0*wit_^2*wit__0^2 - h_0*wit__0^4)*wit__1^2 + 4*(2*wit_^4*wit__0^2 - wit_^2*wit__0^4 + 2*h_0*wit_^2*wit__1^2)*wit__11^2)*wit__12^2 + 4*(3*wit_^2*wit__0^2*wit__1^2*wit__11^4 + wit__0^2*wit__1^2*wit__11^6 + 2*h_0*wit_^2*wit__0^2*wit__1^4 + (2*wit_^4*wit__0^4 - wit_^2*wit__0^6)*wit__1^2 + (2*h_0*wit__0^2*wit__1^4 + (2*wit_^4*wit__0^2 + 2*wit_^2*wit__0^4 - wit__0^6)*wit__1^2)*wit__11^2)*wit__13^2 - 4*((2*wit_^2*wit__1^2*wit__11^2 + wit__1^2*wit__11^4 + 2*h_0*wit__1^4 + (2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__12^2 + 2*(wit_^2*wit__0^2*wit__1^4 + wit__0^2*wit__1^4*wit__11^2)*wit__13^2)*wit__2^2 + (4*wit_^4*wit__0^4 - 4*wit_^2*wit__0^6 + wit__0^8 + 4*wit_^2*wit__11^6 + wit__11^8 + 4*wit__1^4*wit__2^4 + 4*h_0^2*wit__1^4 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4 + 2*h_0*wit__1^2)*wit__11^4 + 4*(2*h_0*wit_^2*wit__0^2 - h_0*wit__0^4)*wit__1^2 + 4*(2*wit_^4*wit__0^2 - wit_^2*wit__0^4 + 2*h_0*wit_^2*wit__1^2)*wit__11^2 - 4*(2*wit_^2*wit__1^2*wit__11^2 + wit__1^2*wit__11^4 + 2*h_0*wit__1^4 + (2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__2^2)*wit__3^2)*wit__9^4 - 1/2*(2*wit__1^6*wit__12^2*wit__13^4 - wit__1^2*wit__12^6)*wit__2^2 + 1/4*((8*wit_^4 - 8*wit_^2*wit__0^2 + 3*wit__0^4 - wit__11^4 - 6*h_0*wit__1^2 - 2*(wit_^2 + 2*wit__0^2)*wit__11^2)*wit__12^4 + 2*(wit__1^2*wit__11^4 + 2*(3*wit_^2 - wit__0^2)*wit__1^2*wit__11^2 + 2*(2*wit_^4 - wit_^2*wit__0^2)*wit__1^2)*wit__12^2*wit__13^2 + (8*wit_^2*wit__1^4*wit__11^2 + 3*wit__1^4*wit__11^4 + 4*h_0*wit__1^6 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^4)*wit__13^4 - 2*(2*wit__1^6*wit__13^4 - 3*wit__1^2*wit__12^4)*wit__2^2)*wit__3^2 + 3/2*(2*wit_^2*wit__0^2*wit__1^4*wit__12^2*wit__13^2 - wit_^2*wit__0^2*wit__1^2*wit__12^2*wit__3^2 - wit_^2*wit__1^2*wit__11^2*wit__3^4 + (wit_^2*wit__0^2*wit__1^2 + wit_^2*wit__1^2*wit__11^2)*wit__12^4)*wit__5^2 + 1/4*(2*wit__12^6 + 3*wit__12^4*wit__3^2 - wit__3^6 - 3*(wit__1^2*wit__12^4 + 2*wit__1^2*wit__12^2*wit__3^2 + wit__1^2*wit__3^4)*wit__5^2)*wit__6^2 + 1/4*((4*wit_^4*wit__0^2 - 2*wit_^2*wit__0^4 + wit__0^6 - 2*wit__0^2*wit__11^4 + 2*(2*h_0*wit_^2 - (t_0*wit_^2 + h_0)*wit__0^2)*wit__1^2 - 2*(t_0*wit_^2*wit__1^2 - 4*wit_^4 + wit_^2*wit__0^2)*wit__11^2)*wit__12^4 + 4*(2*wit_^4*wit__0^2*wit__1^2 + 3*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 - (t_0*wit_^2*wit__0^2 - 2*h_0*wit_^2)*wit__1^4 + (2*h_0*wit__1^4 + (2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__11^2)*wit__12^2*wit__13^2 + 2*(6*wit_^2*wit__0^2*wit__1^4*wit__11^2 + 3*wit__0^2*wit__1^4*wit__11^4 + 2*h_0*wit__0^2*wit__1^6 + (2*wit_^4*wit__0^2 + 2*wit_^2*wit__0^4 - wit__0^6)*wit__1^4)*wit__13^4 + (4*wit_^4*wit__0^2 - 4*wit_^2*wit__0^4 + wit__0^6 + 2*(4*wit_^2 - wit__0^2)*wit__11^4 - 2*(2*wit_^2 - wit__0^2)*wit__1^2*wit__2^2 + 2*(2*h_0*wit_^2 - h_0*wit__0^2)*wit__1^2 + 2*(t_0*wit_^2*wit__1^2 + 4*wit_^4 - 3*wit_^2*wit__0^2)*wit__11^2)*wit__3^4 - 2*(2*wit__0^2*wit__1^6*wit__13^4 + (2*wit_^2 - wit__0^2)*wit__1^2*wit__12^4 + 4*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__12^2*wit__13^2)*wit__2^2 + 2*((4*wit_^4*wit__0^2 - 5*wit_^2*wit__0^4 + wit__0^6 + 2*(2*wit_^2 - wit__0^2)*wit__11^4 + (4*h_0*wit_^2 + (t_0*wit_^2 - 2*h_0)*wit__0^2)*wit__1^2 + 4*(2*wit_^4 - wit_^2*wit__0^2)*wit__11^2)*wit__12^2 + 2*(3*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 + 2*h_0*wit_^2*wit__1^4 + (2*wit_^4*wit__0^2 - wit_^2*wit__0^4)*wit__1^2 + (2*h_0*wit__1^4 + (2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__11^2)*wit__13^2 - 2*((2*wit_^2 - wit__0^2)*wit__1^2*wit__12^2 + 2*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__13^2)*wit__2^2)*wit__3^2 + (4*wit_^2*wit__1^2*wit__12^2*wit__13^2 + (t_0*wit__1^2 + 4*wit_^2 - 2*wit__0^2 + 2*wit__11^2)*wit__12^4 + 2*(t_0*wit__1^2 + wit_^2 - 2*wit__0^2 + 2*wit__11^2)*wit__12^2*wit__3^2 + (t_0*wit__1^2 + 2*wit_^2 - 2*wit__0^2 + 2*wit__11^2)*wit__3^4)*wit__6^2)*wit__9^2"),(1,"2*wit_^2*wit__0^2*wit__1^4*wit__12^2*wit__13^2 - wit_^2*wit__0^2*wit__1^2*wit__12^2*wit__3^2 - wit_^2*wit__1^2*wit__11^2*wit__3^4 + (wit_^2*wit__0^2*wit__1^2 + wit_^2*wit__1^2*wit__11^2)*wit__12^4 - 1/2*(wit__1^2*wit__12^4 + 2*wit__1^2*wit__12^2*wit__3^2 + wit__1^2*wit__3^4)*wit__6^2"),(2,"1/4*wit__12^8 - wit__12^4*wit__3^4 - wit__12^2*wit__3^6 - 1/4*wit__3^8 - 1/2*(2*wit_^2*wit__0^2*wit__1^2*wit__12^2*wit__13^2 + wit__1^2*wit__12^4*wit__2^2 + (wit_^2*wit__0^2 + wit_^2*wit__11^2 - h_0*wit__1^2)*wit__12^4 - (wit_^2*wit__11^2 - wit__1^2*wit__2^2 + h_0*wit__1^2)*wit__3^4 + (2*wit__1^2*wit__12^2*wit__2^2 - (wit_^2*wit__0^2 + 2*h_0*wit__1^2)*wit__12^2)*wit__3^2)*wit__9^4 + 1/2*(wit_^2*wit__1^2*wit__12^4*wit__13^2 + wit_^2*wit__3^6 + (wit_^2*wit__1^2*wit__13^2 + 2*wit_^2*wit__12^2)*wit__3^4 + (2*wit_^2*wit__1^2*wit__12^2*wit__13^2 + wit_^2*wit__12^4)*wit__3^2)*wit__9^2"),(3,"0"),(4,"1/2*wit__0^2*wit__12^8 - 2*wit__0^2*wit__12^4*wit__3^4 - 2*wit__0^2*wit__12^2*wit__3^6 - 1/2*wit__0^2*wit__3^8 - 1/4*(t_0^3*wit__0^4*wit__1^6 + 4*wit_^4*wit__0^6 - 4*wit_^2*wit__0^8 + wit__0^10 + wit__0^4*wit__11^6 - (3*t_0^2*wit__0^6 - 4*(t_0^2*wit_^2 + h_0*t_0)*wit__0^4)*wit__1^4 + (t_0*wit__0^4*wit__1^2 + 4*wit_^2*wit__0^4 - wit__0^6)*wit__11^4 + (t_0*wit__0^8 - 4*(t_0*wit_^2 + h_0)*wit__0^6 + 4*(t_0*wit_^4 + 2*h_0*wit_^2)*wit__0^4)*wit__1^2 + (t_0^2*wit__0^4*wit__1^4 + 4*wit_^4*wit__0^4 - wit__0^8 - 2*(t_0*wit__0^6 - 2*(t_0*wit_^2 + h_0)*wit__0^4)*wit__1^2)*wit__11^2 - 4*(t_0*wit__0^4*wit__1^4 + wit__0^4*wit__1^2*wit__11^2 + (2*wit_^2*wit__0^4 - wit__0^6)*wit__1^2)*wit__2^2)*wit__9^8 + (2*wit__0^4*wit__1^4*wit__13^2*wit__2^2 - (t_0^2*wit__0^4*wit__1^6 + wit__0^4*wit__1^2*wit__11^4 - (2*t_0*wit__0^6 - (3*t_0*wit_^2 + 2*h_0)*wit__0^4)*wit__1^4 + (2*wit_^4*wit__0^4 - wit_^2*wit__0^6)*wit__1^2 + (t_0*wit__0^4*wit__1^4 + (3*wit_^2*wit__0^4 - wit__0^6)*wit__1^2)*wit__11^2)*wit__13^2)*wit__9^6 - 1/4*(4*(wit__1^2*wit__11^2 + (wit_^2 + wit__0^2)*wit__1^2)*wit__12^4*wit__2^2 - (4*wit_^2*wit__11^4 + wit__11^6 + 4*(h_0*wit_^2 + h_0*wit__0^2)*wit__1^2 + 4*(2*wit_^4 + h_0*wit__1^2)*wit__11^2)*wit__12^4 + 6*(t_0*wit__0^4*wit__1^6 + wit__0^4*wit__1^4*wit__11^2 + (2*wit_^2*wit__0^4 - wit__0^6)*wit__1^4)*wit__13^4 - (4*wit_^2*wit__11^4 + wit__11^6 + 4*h_0*wit__1^2*wit__11^2 + 4*(h_0*wit_^2 + h_0*wit__0^2)*wit__1^2 - 4*(wit__1^2*wit__11^2 + (wit_^2 + wit__0^2)*wit__1^2)*wit__2^2)*wit__3^4 + 2*(4*(wit__1^2*wit__11^2 + (wit_^2 + wit__0^2)*wit__1^2)*wit__12^2*wit__2^2 - (4*wit_^2*wit__11^4 + wit__11^6 + 4*(h_0*wit_^2 + h_0*wit__0^2)*wit__1^2 + 4*(wit_^4 + h_0*wit__1^2)*wit__11^2)*wit__12^2)*wit__3^2 + 2*(2*wit__0^2*wit__12^2*wit__3^2 - (wit_^2 - wit__0^2)*wit__12^4 + (wit_^2 + wit__0^2)*wit__3^4)*wit__6^2 + 4*(wit__1^2*wit__12^4*wit__2^2 + (wit_^2*wit__11^2 - h_0*wit__1^2)*wit__12^4 - (wit_^2*wit__11^2 - wit__1^2*wit__2^2 + h_0*wit__1^2)*wit__3^4 + 2*(wit__1^2*wit__12^2*wit__2^2 - h_0*wit__1^2*wit__12^2)*wit__3^2)*wit__7^2)*wit__9^4 + 1/2*(wit__12^8 - 4*wit__12^4*wit__3^4 - 4*wit__12^2*wit__3^6 - wit__3^8)*wit__7^2 - 1/4*(3*wit__0^4*wit__1^6*wit__13^6 - 8*wit_^2*wit__0^2*wit__1^2*wit__12^2*wit__4^2 - (8*wit_^4 - 10*wit_^2*wit__0^2 + wit__0^4 - wit__11^4 - 4*h_0*wit__1^2 - (4*wit_^2 + wit__0^2)*wit__11^2)*wit__12^6 - (3*wit__1^2*wit__11^4 + 4*h_0*wit__1^4 + 2*(4*wit_^2 - wit__0^2)*wit__1^2*wit__11^2 + (4*wit_^4 - 4*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__12^4*wit__13^2 - (3*wit__0^2*wit__1^4*wit__11^2 + (4*wit_^2*wit__0^2 - wit__0^4)*wit__1^4)*wit__12^2*wit__13^4 - 2*(2*wit_^4 + wit_^2*wit__0^2 + (4*wit_^2 - wit__0^2)*wit__11^2)*wit__3^6 - ((8*wit_^4 + 2*wit_^2*wit__0^2 + 2*wit__0^4 - wit__11^4 - 4*h_0*wit__1^2 + 6*(2*wit_^2 - wit__0^2)*wit__11^2)*wit__12^2 + (8*wit_^2*wit__1^2*wit__11^2 + 3*wit__1^2*wit__11^4 + 4*h_0*wit__1^4 + 2*(2*wit_^4 + 4*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__13^2 - 4*(wit__1^4*wit__13^2 - wit__1^2*wit__12^2)*wit__2^2)*wit__3^4 + 4*(wit__1^4*wit__12^4*wit__13^2 - wit__1^2*wit__12^6)*wit__2^2 - ((12*wit_^4 - 6*wit_^2*wit__0^2 + 4*wit__0^4 - 5*wit__0^2*wit__11^2 - 2*wit__11^4 - 8*h_0*wit__1^2)*wit__12^4 + 2*(3*wit__1^2*wit__11^4 + 4*h_0*wit__1^4 + (8*wit_^2 - wit__0^2)*wit__1^2*wit__11^2 + 2*(2*wit_^4 + wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__12^2*wit__13^2 + (4*wit_^2*wit__0^2*wit__1^4 + 3*wit__0^2*wit__1^4*wit__11^2)*wit__13^4 - 8*(wit__1^4*wit__12^2*wit__13^2 - wit__1^2*wit__12^4)*wit__2^2)*wit__3^2 - 2*(wit__12^6 + 3*wit__12^4*wit__3^2 + 3*wit__12^2*wit__3^4 + wit__3^6)*wit__6^2 + 4*(wit_^2*wit__1^2*wit__12^4*wit__13^2 + wit_^2*wit__12^6 - wit_^2*wit__12^4*wit__3^2 - wit_^2*wit__3^6 - (wit_^2*wit__1^2*wit__13^2 + 3*wit_^2*wit__12^2)*wit__3^4)*wit__7^2)*wit__9^2"),(5,"-1/2*(2*wit_^4 - 2*wit_^2*wit__0^2 - h_0*wit__1^2)*wit__12^6 - 1/4*(wit__0^2*wit__1^2*wit__11^2 + wit__1^2*wit__11^4 + 4*h_0*wit__1^4 - (2*wit_^2*wit__0^2 + wit__0^4)*wit__1^2)*wit__12^4*wit__13^2 + 1/4*(3*wit__1^4*wit__11^4 + 4*h_0*wit__1^6 + 2*(4*wit_^2 - wit__0^2)*wit__1^4*wit__11^2 + (4*wit_^4 - wit__0^4)*wit__1^4)*wit__12^2*wit__13^4 + 1/4*(3*wit__0^2*wit__1^6*wit__11^2 + (4*wit_^2*wit__0^2 - wit__0^4)*wit__1^6)*wit__13^6 - 1/4*(4*wit_^4 - wit__0^4 - 2*wit__1^2*wit__2^2 + 2*h_0*wit__1^2 + 2*(6*wit_^2 - wit__0^2)*wit__11^2)*wit__3^6 + 1/4*(4*wit_^4*wit__0^6 - 4*wit_^2*wit__0^8 + wit__0^10 + 4*wit_^2*wit__0^2*wit__11^6 + wit__0^2*wit__11^8 + 4*wit__0^2*wit__1^4*wit__2^4 + 4*h_0^2*wit__0^2*wit__1^4 + 2*(2*wit_^4*wit__0^2 + 2*wit_^2*wit__0^4 - wit__0^6 + 2*h_0*wit__0^2*wit__1^2)*wit__11^4 + 4*(2*h_0*wit_^2*wit__0^4 - h_0*wit__0^6)*wit__1^2 + 4*(2*wit_^4*wit__0^4 - wit_^2*wit__0^6 + 2*h_0*wit_^2*wit__0^2*wit__1^2)*wit__11^2 - 4*(2*wit_^2*wit__0^2*wit__1^2*wit__11^2 + wit__0^2*wit__1^2*wit__11^4 + 2*h_0*wit__0^2*wit__1^4 + (2*wit_^2*wit__0^4 - wit__0^6)*wit__1^2)*wit__2^2)*wit__9^6 - 1/4*((4*wit_^4 + 4*wit_^2*wit__0^2 - wit__0^4 - wit__11^4 + 2*h_0*wit__1^2 + 4*(2*wit_^2 - wit__0^2)*wit__11^2)*wit__12^2 + (3*wit__1^2*wit__11^4 + 4*h_0*wit__1^4 + 2*(2*wit_^2 + wit__0^2)*wit__1^2*wit__11^2 + 2*(3*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__13^2 - 2*(2*wit__1^4*wit__13^2 + wit__1^2*wit__12^2)*wit__2^2)*wit__3^4 - 2*(wit_^2*wit__0^2*wit__1^2 + wit_^2*wit__1^2*wit__11^2)*wit__12^2*wit__4^2 + 1/4*(4*wit__1^4*wit__12^2*wit__2^4 + (4*wit_^4*wit__0^4 - 4*wit_^2*wit__0^6 + wit__0^8 + 4*wit_^2*wit__11^6 + wit__11^8 + 4*h_0^2*wit__1^4 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4 + 2*h_0*wit__1^2)*wit__11^4 + 4*(2*h_0*wit_^2*wit__0^2 - h_0*wit__0^4)*wit__1^2 + 4*(2*wit_^4*wit__0^2 - wit_^2*wit__0^4 + 2*h_0*wit_^2*wit__1^2)*wit__11^2)*wit__12^2 + 4*(3*wit_^2*wit__0^2*wit__1^2*wit__11^4 + wit__0^2*wit__1^2*wit__11^6 + 2*h_0*wit_^2*wit__0^2*wit__1^4 + (2*wit_^4*wit__0^4 - wit_^2*wit__0^6)*wit__1^2 + (2*h_0*wit__0^2*wit__1^4 + (2*wit_^4*wit__0^2 + 2*wit_^2*wit__0^4 - wit__0^6)*wit__1^2)*wit__11^2)*wit__13^2 - 4*((2*wit_^2*wit__1^2*wit__11^2 + wit__1^2*wit__11^4 + 2*h_0*wit__1^4 + (2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__12^2 + 2*(wit_^2*wit__0^2*wit__1^4 + wit__0^2*wit__1^4*wit__11^2)*wit__13^2)*wit__2^2 + (4*wit_^4*wit__0^4 - 4*wit_^2*wit__0^6 + wit__0^8 + 4*wit_^2*wit__11^6 + wit__11^8 + 4*wit__1^4*wit__2^4 + 4*h_0^2*wit__1^4 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4 + 2*h_0*wit__1^2)*wit__11^4 + 4*(2*h_0*wit_^2*wit__0^2 - h_0*wit__0^4)*wit__1^2 + 4*(2*wit_^4*wit__0^2 - wit_^2*wit__0^4 + 2*h_0*wit_^2*wit__1^2)*wit__11^2 - 4*(2*wit_^2*wit__1^2*wit__11^2 + wit__1^2*wit__11^4 + 2*h_0*wit__1^4 + (2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__2^2)*wit__3^2)*wit__9^4 - 1/2*(2*wit__1^6*wit__12^2*wit__13^4 - 2*wit__1^4*wit__12^4*wit__13^2 + wit__1^2*wit__12^6)*wit__2^2 - 1/4*((4*wit_^4 - 4*wit_^2*wit__0^2 + wit__0^4 - wit__0^2*wit__11^2 - wit__11^4 - 2*h_0*wit__1^2)*wit__12^4 + 2*(2*wit__1^2*wit__11^4 + 4*h_0*wit__1^4 + (2*wit_^2 + wit__0^2)*wit__1^2*wit__11^2 - 2*(wit_^2*wit__0^2 + wit__0^4)*wit__1^2)*wit__12^2*wit__13^2 - (3*wit__1^4*wit__11^4 + 4*h_0*wit__1^6 + (8*wit_^2 - 3*wit__0^2)*wit__1^4*wit__11^2 + 2*(2*wit_^4 - wit__0^4)*wit__1^4)*wit__13^4 + 2*(2*wit__1^6*wit__13^4 - 4*wit__1^4*wit__12^2*wit__13^2 + wit__1^2*wit__12^4)*wit__2^2)*wit__3^2 + (2*wit_^2*wit__0^2*wit__1^4*wit__12^2*wit__13^2 - wit_^2*wit__0^2*wit__1^2*wit__12^2*wit__3^2 - wit_^2*wit__1^2*wit__11^2*wit__3^4 + (wit_^2*wit__0^2*wit__1^2 + wit_^2*wit__1^2*wit__11^2)*wit__12^4)*wit__5^2 - 1/2*(2*wit__12^4*wit__3^2 + 4*wit__12^2*wit__3^4 + 2*wit__3^6 + (wit__1^2*wit__12^4 + 2*wit__1^2*wit__12^2*wit__3^2 + wit__1^2*wit__3^4)*wit__5^2)*wit__6^2 + (2*wit_^2*wit__1^2*wit__12^4*wit__13^2 + 2*wit_^2*wit__1^2*wit__12^2*wit__13^2*wit__3^2 + wit_^2*wit__12^6 - wit_^2*wit__12^2*wit__3^4 + (2*wit_^2*wit__1^4*wit__12^2*wit__13^2 + wit_^2*wit__1^2*wit__12^4 - wit_^2*wit__1^2*wit__12^2*wit__3^2)*wit__5^2)*wit__7^2 + 1/4*((4*wit_^4*wit__0^2 - 6*wit_^2*wit__0^4 + wit__0^6 - wit__11^6 - 2*h_0*wit__0^2*wit__1^2 - 2*(2*wit_^2 + wit__0^2)*wit__11^4 - 2*(3*wit_^2*wit__0^2 + 2*h_0*wit__1^2)*wit__11^2)*wit__12^4 + 4*(3*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 + 2*h_0*wit_^2*wit__1^4 + 2*(wit_^4*wit__0^2 - wit_^2*wit__0^4)*wit__1^2 + (2*h_0*wit__1^4 + (2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__11^2)*wit__12^2*wit__13^2 + 2*(6*wit_^2*wit__0^2*wit__1^4*wit__11^2 + 3*wit__0^2*wit__1^4*wit__11^4 + 2*h_0*wit__0^2*wit__1^6 + (2*wit_^4*wit__0^2 + 2*wit_^2*wit__0^4 - wit__0^6)*wit__1^4)*wit__13^4 + (4*wit_^4*wit__0^2 - 4*wit_^2*wit__0^4 + wit__0^6 - wit__11^6 - 2*h_0*wit__0^2*wit__1^2 + 2*(2*wit_^2 - wit__0^2)*wit__11^4 + 2*(4*wit_^4 - wit_^2*wit__0^2 - 2*h_0*wit__1^2)*wit__11^2 + 2*(wit__0^2*wit__1^2 + 2*wit__1^2*wit__11^2)*wit__2^2)*wit__3^4 - 2*(2*wit__0^2*wit__1^6*wit__13^4 - (wit__0^2*wit__1^2 + 2*wit__1^2*wit__11^2)*wit__12^4 + 4*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__12^2*wit__13^2)*wit__2^2 + 2*((4*wit_^4*wit__0^2 - 3*wit_^2*wit__0^4 + wit__0^6 - 2*wit__0^2*wit__11^4 - wit__11^6 - 2*h_0*wit__0^2*wit__1^2 + 4*(wit_^4 - wit_^2*wit__0^2 - h_0*wit__1^2)*wit__11^2)*wit__12^2 + 2*(3*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 + 2*h_0*wit_^2*wit__1^4 + (2*wit_^4*wit__0^2 - wit_^2*wit__0^4)*wit__1^2 + (2*h_0*wit__1^4 + (2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__11^2)*wit__13^2 + 2*((wit__0^2*wit__1^2 + 2*wit__1^2*wit__11^2)*wit__12^2 - 2*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__13^2)*wit__2^2)*wit__3^2 + 2*(2*wit_^2*wit__1^2*wit__12^2*wit__13^2 + (wit_^2 + wit__11^2)*wit__12^4 + (wit_^2 + 2*wit__11^2)*wit__12^2*wit__3^2 + (2*wit_^2 + wit__11^2)*wit__3^4)*wit__6^2 + 2*(2*wit_^2*wit__11^2*wit__12^4 - 2*wit_^2*wit__11^2*wit__3^4 - (wit__12^4 + 2*wit__12^2*wit__3^2 + wit__3^4)*wit__6^2)*wit__7^2)*wit__9^2"),(6,"-1/4*wit__0^4*wit__1^8*wit__13^4*wit__5^2 + 1/4*wit__0^4*wit__1^8*wit__13^2*wit__5^4 - 1/4*wit__0^4*wit__1^8*wit__5^6 - 2*wit_^2*wit__0^2*wit__1^4*wit__12^2*wit__4^2 + 1/4*(wit__0^2*wit__1^2*wit__11^2 + wit__1^2*wit__11^4 + 4*h_0*wit__1^4 - (8*wit_^4 - 6*wit_^2*wit__0^2 + wit__0^4)*wit__1^2)*wit__12^6 - 1/4*(3*wit__1^4*wit__11^4 + 4*h_0*wit__1^6 + 2*(4*wit_^2 - wit__0^2)*wit__1^4*wit__11^2 + (4*wit_^4 - wit__0^4)*wit__1^4)*wit__12^4*wit__13^2 - 1/4*(3*wit__0^2*wit__1^6*wit__11^2 + (4*wit_^2*wit__0^2 - wit__0^4)*wit__1^6)*wit__12^2*wit__13^4 - 1/2*((2*wit_^2 - wit__0^2)*wit__1^2*wit__11^2 + (2*wit_^4 - wit_^2*wit__0^2)*wit__1^2)*wit__3^6 + (t_0^3*wit__0^4*wit__1^8 - 2*t_0*wit__0^4*wit__1^6*wit__2^2 + 2*t_0*wit__0^4*wit__1^4*wit__7^4 + 2*h_0*t_0*wit__0^4*wit__1^6 + (3*t_0^2*wit__0^4*wit__1^6 - 2*wit__0^4*wit__1^4*wit__2^2 + 2*h_0*wit__0^4*wit__1^4)*wit__7^2)*wit__9^6 + 1/4*((wit__1^2*wit__11^4 + 4*h_0*wit__1^4 - 2*(4*wit_^2 - 3*wit__0^2)*wit__1^2*wit__11^2 - 2*(4*wit_^4 - 5*wit_^2*wit__0^2 + wit__0^4)*wit__1^2)*wit__12^2 - (8*wit_^2*wit__1^4*wit__11^2 + 3*wit__1^4*wit__11^4 + 4*h_0*wit__1^6 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^4)*wit__13^2 + 4*(wit__1^6*wit__13^2 - wit__1^4*wit__12^2)*wit__2^2)*wit__3^4 - 1/2*(2*wit__0^4*wit__1^6*wit__13^2*wit__2^2 - 2*(wit__0^4*wit__1^4*wit__13^2 - wit__0^4*wit__1^4*wit__5^2)*wit__7^4 - (3*t_0^2*wit__0^4*wit__1^8 + 2*h_0*wit__0^4*wit__1^6)*wit__13^2 + (3*t_0^2*wit__0^4*wit__1^8 - 2*wit__0^4*wit__1^6*wit__2^2 + 2*h_0*wit__0^4*wit__1^6)*wit__5^2 - 6*(t_0*wit__0^4*wit__1^6*wit__13^2 - t_0*wit__0^4*wit__1^6*wit__5^2)*wit__7^2)*wit__9^4 + (wit__1^6*wit__12^4*wit__13^2 - wit__1^4*wit__12^6)*wit__2^2 + 1/4*((2*wit__1^2*wit__11^4 + 8*h_0*wit__1^4 - (4*wit_^2 - 5*wit__0^2)*wit__1^2*wit__11^2 - 2*(6*wit_^4 - 5*wit_^2*wit__0^2 + 2*wit__0^4)*wit__1^2)*wit__12^4 - 2*(3*wit__1^4*wit__11^4 + 4*h_0*wit__1^6 + (8*wit_^2 - wit__0^2)*wit__1^4*wit__11^2 + 2*(2*wit_^4 + wit_^2*wit__0^2 - wit__0^4)*wit__1^4)*wit__12^2*wit__13^2 - (4*wit_^2*wit__0^2*wit__1^6 + 3*wit__0^2*wit__1^6*wit__11^2)*wit__13^4 + 8*(wit__1^6*wit__12^2*wit__13^2 - wit__1^4*wit__12^4)*wit__2^2)*wit__3^2 + 1/4*(4*t_0*wit__0^4*wit__1^8*wit__13^4 - 4*t_0*wit__0^4*wit__1^8*wit__13^2*wit__5^2 + 4*t_0*wit__0^4*wit__1^8*wit__5^4 + 4*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__12^4*wit__2^2 - (4*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 + 4*h_0*wit_^2*wit__1^4 + 4*(2*wit_^4*wit__1^2 + h_0*wit__1^4)*wit__11^2)*wit__12^4 - (4*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 + 4*h_0*wit_^2*wit__1^4 + 4*h_0*wit__1^4*wit__11^2 - 4*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__2^2)*wit__3^4 + 2*(4*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__12^2*wit__2^2 - (4*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 + 4*h_0*wit_^2*wit__1^4 + 4*(wit_^4*wit__1^2 + h_0*wit__1^4)*wit__11^2)*wit__12^2)*wit__3^2 - 2*(wit_^2*wit__1^2*wit__12^4 - wit_^2*wit__1^2*wit__3^4)*wit__6^2 + 4*(wit__0^4*wit__1^6*wit__13^4 - wit__0^4*wit__1^6*wit__13^2*wit__5^2 + wit__0^4*wit__1^6*wit__5^4)*wit__7^2)*wit__9^2"),(7,"-2*wit_^2*wit__1^2*wit__12^6*wit__13^2 - 2*wit_^2*wit__1^2*wit__12^4*wit__13^2*wit__3^2 - 1/4*(4*wit_^2 - wit__0^2 - 2*wit__11^2)*wit__12^8 + (wit_^2 - 2*wit__0^2 - 2*wit__11^2)*wit__12^4*wit__3^4 - 2*(wit__0^2 + wit__11^2)*wit__12^2*wit__3^6 - 1/2*(wit__0^2 + wit__11^2)*wit__3^8 - (t_0^3*wit__0^4*wit__1^6 + t_0^2*wit__0^4*wit__1^4*wit__7^2 - 2*t_0*wit__0^4*wit__1^4*wit__2^2 - (t_0^2*wit__0^6 - (t_0^2*wit_^2 + 2*h_0*t_0)*wit__0^4)*wit__1^4)*wit__9^8 - (2*t_0*wit__0^4*wit__1^4*wit__13^2*wit__7^2 - 2*wit__0^4*wit__1^4*wit__13^2*wit__2^2 + (3*t_0^2*wit__0^4*wit__1^6 - 2*(t_0*wit__0^6 - (t_0*wit_^2 + h_0)*wit__0^4)*wit__1^4)*wit__13^2)*wit__9^6 - 1/2*(2*wit__0^4*wit__1^4*wit__13^4*wit__7^2 + 2*(wit__0^2*wit__1^2 + wit__1^2*wit__11^2)*wit__12^4*wit__2^2 + 2*(wit_^2*wit__11^4 - h_0*wit__0^2*wit__1^2 - h_0*wit__1^2*wit__11^2)*wit__12^4 + 2*(3*t_0*wit__0^4*wit__1^6 + (wit_^2*wit__0^4 - wit__0^6)*wit__1^4)*wit__13^4 - 2*(wit_^2*wit__11^4 + h_0*wit__0^2*wit__1^2 + h_0*wit__1^2*wit__11^2 - (wit__0^2*wit__1^2 + wit__1^2*wit__11^2)*wit__2^2)*wit__3^4 + 4*((wit__0^2*wit__1^2 + wit__1^2*wit__11^2)*wit__12^2*wit__2^2 - (h_0*wit__0^2*wit__1^2 + h_0*wit__1^2*wit__11^2)*wit__12^2)*wit__3^2 + (wit__0^2*wit__12^4 + 2*wit__0^2*wit__12^2*wit__3^2 + wit__0^2*wit__3^4)*wit__6^2)*wit__9^4 - 1/2*(2*wit__0^4*wit__1^6*wit__13^6 + 4*wit_^2*wit__11^2*wit__12^6 - 2*(wit_^2*wit__0^2*wit__1^2 - wit_^2*wit__1^2*wit__11^2)*wit__12^4*wit__13^2 - 2*(wit_^2*wit__0^2 + wit_^2*wit__11^2)*wit__3^6 - 2*(2*(wit_^2*wit__0^2 + 2*wit_^2*wit__11^2)*wit__12^2 + (wit_^2*wit__0^2*wit__1^2 + wit_^2*wit__1^2*wit__11^2)*wit__13^2)*wit__3^4 - 2*(2*wit_^2*wit__0^2*wit__1^2*wit__12^2*wit__13^2 + (wit_^2*wit__0^2 + wit_^2*wit__11^2)*wit__12^4)*wit__3^2 - (wit__12^6 + 2*wit__12^4*wit__3^2 + wit__12^2*wit__3^4)*wit__6^2)*wit__9^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(0,1,4),List((0,"w_0","1","1"),(0,"abs_1","-rv_0*t_0 + r_0","1"),(7,"r_0","rv_0*t_0 - wit__5^2 + rp_0","-1"),(10,"rv_0","wit__9^2","1"),(0,"min_1","0","1"),(0,"dhd_0","dhf_0 - max_1","1"),(6,"dhf_0","wit__7^2 + max_1","-1"),(0,"max_1","wit__0^2","-1"),(0,"a_0","wit__1^2","1"),(0,"hp_0","wit__2^2","1"),(3,"rp_0","wit__5^2 + wit__8^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__1*wit__3^2*wit__4*wit__11*wit_ + wit__1*wit__4*wit__11*wit__12^2*wit_"),("1","wit__1*wit__3*wit__4*wit__0*wit__12*wit_"),("1","wit__1*wit__4*wit__0*wit__12^2*wit_"),("1/2","wit__3^2*wit__11*wit__12^2*wit_ + wit__11*wit__12^4*wit_"),("1/2","wit__3*wit__0*wit__12^3*wit_"),("1/4","wit__0*wit__12^4*wit_"),("1/2","wit__1*wit__3^2*wit__4*wit__6 + wit__1*wit__4*wit__6*wit__12^2"),("1/4","wit__3^2*wit__6*wit__12^2 + wit__6*wit__12^4"),("1/4","wit__7*wit__0*wit__12^4")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"-2*wit_^2*wit__1^2*wit__11^2*wit__12^2*wit__4^2 + 1/4*(4*wit_^4 - 6*wit_^2*wit__0^2 + wit__0^4 - wit__11^4 + 2*h_0*wit__1^2 - (4*wit_^2 + wit__0^2)*wit__11^2)*wit__12^6 + 1/4*(2*wit__1^2*wit__11^4 + (8*wit_^2 - 3*wit__0^2)*wit__1^2*wit__11^2 + 2*(2*wit_^4 - 3*wit_^2*wit__0^2)*wit__1^2)*wit__12^4*wit__13^2 + 1/4*(3*wit__1^4*wit__11^4 - 4*h_0*wit__1^6 + (8*wit_^2 + wit__0^2)*wit__1^4*wit__11^2 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^4)*wit__12^2*wit__13^4 + 1/4*(4*wit_^2*wit__0^2*wit__1^6 + 3*wit__0^2*wit__1^6*wit__11^2)*wit__13^6 - 1/4*(2*wit_^2*wit__0^2 - wit__0^4 + 2*wit_^2*wit__11^2 - 2*wit__1^2*wit__2^2 - 2*h_0*wit__1^2)*wit__3^6 + 1/4*(4*wit_^4*wit__0^6 - 4*wit_^2*wit__0^8 + wit__0^10 + 4*wit_^2*wit__0^2*wit__11^6 + wit__0^2*wit__11^8 + 4*wit__0^2*wit__1^4*wit__2^4 + 4*h_0^2*wit__0^2*wit__1^4 + 2*(2*wit_^4*wit__0^2 + 2*wit_^2*wit__0^4 - wit__0^6 - 2*h_0*wit__0^2*wit__1^2)*wit__11^4 - 4*(2*h_0*wit_^2*wit__0^4 - h_0*wit__0^6)*wit__1^2 + 4*(2*wit_^4*wit__0^4 - wit_^2*wit__0^6 - 2*h_0*wit_^2*wit__0^2*wit__1^2)*wit__11^2 - 4*(2*wit_^2*wit__0^2*wit__1^2*wit__11^2 + wit__0^2*wit__1^2*wit__11^4 - 2*h_0*wit__0^2*wit__1^4 + (2*wit_^2*wit__0^4 - wit__0^6)*wit__1^2)*wit__2^2)*wit__9^6 + 1/4*(6*wit__1^2*wit__12^2*wit__2^2 + (4*wit_^4 - 8*wit_^2*wit__0^2 + 3*wit__0^4 + 6*h_0*wit__1^2 + 2*(2*wit_^2 - wit__0^2)*wit__11^2)*wit__12^2 + 2*((2*wit_^2 - wit__0^2)*wit__1^2*wit__11^2 + (2*wit_^4 - wit_^2*wit__0^2)*wit__1^2)*wit__13^2)*wit__3^4 + (2*wit_^2*wit__1^4*wit__12^2*wit__13^2 + wit_^2*wit__1^2*wit__12^4 - wit_^2*wit__1^2*wit__12^2*wit__3^2)*wit__5^2*wit__7^2 + 1/4*(4*wit__1^4*wit__12^2*wit__2^4 + (4*wit_^4*wit__0^4 - 4*wit_^2*wit__0^6 + wit__0^8 + 4*wit_^2*wit__11^6 + wit__11^8 + 4*h_0^2*wit__1^4 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4 - 2*h_0*wit__1^2)*wit__11^4 - 4*(2*h_0*wit_^2*wit__0^2 - h_0*wit__0^4)*wit__1^2 + 4*(2*wit_^4*wit__0^2 - wit_^2*wit__0^4 - 2*h_0*wit_^2*wit__1^2)*wit__11^2)*wit__12^2 + 4*(3*wit_^2*wit__0^2*wit__1^2*wit__11^4 + wit__0^2*wit__1^2*wit__11^6 - 2*h_0*wit_^2*wit__0^2*wit__1^4 + (2*wit_^4*wit__0^4 - wit_^2*wit__0^6)*wit__1^2 - (2*h_0*wit__0^2*wit__1^4 - (2*wit_^4*wit__0^2 + 2*wit_^2*wit__0^4 - wit__0^6)*wit__1^2)*wit__11^2)*wit__13^2 - 4*((2*wit_^2*wit__1^2*wit__11^2 + wit__1^2*wit__11^4 - 2*h_0*wit__1^4 + (2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__12^2 + 2*(wit_^2*wit__0^2*wit__1^4 + wit__0^2*wit__1^4*wit__11^2)*wit__13^2)*wit__2^2 + (4*wit_^4*wit__0^4 - 4*wit_^2*wit__0^6 + wit__0^8 + 4*wit_^2*wit__11^6 + wit__11^8 + 4*wit__1^4*wit__2^4 + 4*h_0^2*wit__1^4 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4 - 2*h_0*wit__1^2)*wit__11^4 - 4*(2*h_0*wit_^2*wit__0^2 - h_0*wit__0^4)*wit__1^2 + 4*(2*wit_^4*wit__0^2 - wit_^2*wit__0^4 - 2*h_0*wit_^2*wit__1^2)*wit__11^2 - 4*(2*wit_^2*wit__1^2*wit__11^2 + wit__1^2*wit__11^4 - 2*h_0*wit__1^4 + (2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__2^2)*wit__3^2)*wit__9^4 - 1/2*(2*wit__1^6*wit__12^2*wit__13^4 - wit__1^2*wit__12^6)*wit__2^2 + 1/4*((8*wit_^4 - 8*wit_^2*wit__0^2 + 3*wit__0^4 - wit__11^4 + 6*h_0*wit__1^2 - 2*(wit_^2 + 2*wit__0^2)*wit__11^2)*wit__12^4 + 2*(wit__1^2*wit__11^4 + 2*(3*wit_^2 - wit__0^2)*wit__1^2*wit__11^2 + 2*(2*wit_^4 - wit_^2*wit__0^2)*wit__1^2)*wit__12^2*wit__13^2 + (8*wit_^2*wit__1^4*wit__11^2 + 3*wit__1^4*wit__11^4 - 4*h_0*wit__1^6 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^4)*wit__13^4 - 2*(2*wit__1^6*wit__13^4 - 3*wit__1^2*wit__12^4)*wit__2^2)*wit__3^2 + 3/2*(2*wit_^2*wit__0^2*wit__1^4*wit__12^2*wit__13^2 - wit_^2*wit__0^2*wit__1^2*wit__12^2*wit__3^2 - wit_^2*wit__1^2*wit__11^2*wit__3^4 + (wit_^2*wit__0^2*wit__1^2 + wit_^2*wit__1^2*wit__11^2)*wit__12^4)*wit__5^2 + 1/4*(2*wit__12^6 + 3*wit__12^4*wit__3^2 - wit__3^6 - 3*(wit__1^2*wit__12^4 + 2*wit__1^2*wit__12^2*wit__3^2 + wit__1^2*wit__3^4)*wit__5^2)*wit__6^2 + 1/4*((4*wit_^4*wit__0^2 - 2*wit_^2*wit__0^4 + wit__0^6 - 2*wit__0^2*wit__11^4 - 2*(2*h_0*wit_^2 + (t_0*wit_^2 - h_0)*wit__0^2)*wit__1^2 - 2*(t_0*wit_^2*wit__1^2 - 4*wit_^4 + wit_^2*wit__0^2)*wit__11^2)*wit__12^4 + 4*(2*wit_^4*wit__0^2*wit__1^2 + 3*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 - (t_0*wit_^2*wit__0^2 + 2*h_0*wit_^2)*wit__1^4 - (2*h_0*wit__1^4 - (2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__11^2)*wit__12^2*wit__13^2 + 2*(6*wit_^2*wit__0^2*wit__1^4*wit__11^2 + 3*wit__0^2*wit__1^4*wit__11^4 - 2*h_0*wit__0^2*wit__1^6 + (2*wit_^4*wit__0^2 + 2*wit_^2*wit__0^4 - wit__0^6)*wit__1^4)*wit__13^4 + (4*wit_^4*wit__0^2 - 4*wit_^2*wit__0^4 + wit__0^6 + 2*(4*wit_^2 - wit__0^2)*wit__11^4 - 2*(2*wit_^2 - wit__0^2)*wit__1^2*wit__2^2 - 2*(2*h_0*wit_^2 - h_0*wit__0^2)*wit__1^2 + 2*(t_0*wit_^2*wit__1^2 + 4*wit_^4 - 3*wit_^2*wit__0^2)*wit__11^2)*wit__3^4 - 2*(2*wit__0^2*wit__1^6*wit__13^4 + (2*wit_^2 - wit__0^2)*wit__1^2*wit__12^4 + 4*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__12^2*wit__13^2)*wit__2^2 + 2*((4*wit_^4*wit__0^2 - 5*wit_^2*wit__0^4 + wit__0^6 + 2*(2*wit_^2 - wit__0^2)*wit__11^4 - (4*h_0*wit_^2 - (t_0*wit_^2 + 2*h_0)*wit__0^2)*wit__1^2 + 4*(2*wit_^4 - wit_^2*wit__0^2)*wit__11^2)*wit__12^2 + 2*(3*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 - 2*h_0*wit_^2*wit__1^4 + (2*wit_^4*wit__0^2 - wit_^2*wit__0^4)*wit__1^2 - (2*h_0*wit__1^4 - (2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__11^2)*wit__13^2 - 2*((2*wit_^2 - wit__0^2)*wit__1^2*wit__12^2 + 2*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__13^2)*wit__2^2)*wit__3^2 + (4*wit_^2*wit__1^2*wit__12^2*wit__13^2 + (t_0*wit__1^2 + 4*wit_^2 - 2*wit__0^2 + 2*wit__11^2)*wit__12^4 + 2*(t_0*wit__1^2 + wit_^2 - 2*wit__0^2 + 2*wit__11^2)*wit__12^2*wit__3^2 + (t_0*wit__1^2 + 2*wit_^2 - 2*wit__0^2 + 2*wit__11^2)*wit__3^4)*wit__6^2)*wit__9^2"),(1,"2*wit_^2*wit__0^2*wit__1^4*wit__12^2*wit__13^2 - wit_^2*wit__0^2*wit__1^2*wit__12^2*wit__3^2 - wit_^2*wit__1^2*wit__11^2*wit__3^4 + (wit_^2*wit__0^2*wit__1^2 + wit_^2*wit__1^2*wit__11^2)*wit__12^4 - 1/2*(wit__1^2*wit__12^4 + 2*wit__1^2*wit__12^2*wit__3^2 + wit__1^2*wit__3^4)*wit__6^2"),(2,"1/4*wit__12^8 - wit__12^4*wit__3^4 - wit__12^2*wit__3^6 - 1/4*wit__3^8 - 1/2*(2*wit_^2*wit__0^2*wit__1^2*wit__12^2*wit__13^2 + wit__1^2*wit__12^4*wit__2^2 + (wit_^2*wit__0^2 + wit_^2*wit__11^2 + h_0*wit__1^2)*wit__12^4 - (wit_^2*wit__11^2 - wit__1^2*wit__2^2 - h_0*wit__1^2)*wit__3^4 + (2*wit__1^2*wit__12^2*wit__2^2 - (wit_^2*wit__0^2 - 2*h_0*wit__1^2)*wit__12^2)*wit__3^2)*wit__9^4 + 1/2*(wit_^2*wit__1^2*wit__12^4*wit__13^2 + wit_^2*wit__3^6 + (wit_^2*wit__1^2*wit__13^2 + 2*wit_^2*wit__12^2)*wit__3^4 + (2*wit_^2*wit__1^2*wit__12^2*wit__13^2 + wit_^2*wit__12^4)*wit__3^2)*wit__9^2"),(3,"0"),(4,"1/2*wit__0^2*wit__12^8 - 2*wit__0^2*wit__12^4*wit__3^4 - 2*wit__0^2*wit__12^2*wit__3^6 - 1/2*wit__0^2*wit__3^8 - 1/4*(t_0^3*wit__0^4*wit__1^6 + 4*wit_^4*wit__0^6 - 4*wit_^2*wit__0^8 + wit__0^10 + wit__0^4*wit__11^6 - (3*t_0^2*wit__0^6 - 4*(t_0^2*wit_^2 - h_0*t_0)*wit__0^4)*wit__1^4 + (t_0*wit__0^4*wit__1^2 + 4*wit_^2*wit__0^4 - wit__0^6)*wit__11^4 + (t_0*wit__0^8 - 4*(t_0*wit_^2 - h_0)*wit__0^6 + 4*(t_0*wit_^4 - 2*h_0*wit_^2)*wit__0^4)*wit__1^2 + (t_0^2*wit__0^4*wit__1^4 + 4*wit_^4*wit__0^4 - wit__0^8 - 2*(t_0*wit__0^6 - 2*(t_0*wit_^2 - h_0)*wit__0^4)*wit__1^2)*wit__11^2 - 4*(t_0*wit__0^4*wit__1^4 + wit__0^4*wit__1^2*wit__11^2 + (2*wit_^2*wit__0^4 - wit__0^6)*wit__1^2)*wit__2^2)*wit__9^8 + (2*wit__0^4*wit__1^4*wit__13^2*wit__2^2 - (t_0^2*wit__0^4*wit__1^6 + wit__0^4*wit__1^2*wit__11^4 - (2*t_0*wit__0^6 - (3*t_0*wit_^2 - 2*h_0)*wit__0^4)*wit__1^4 + (2*wit_^4*wit__0^4 - wit_^2*wit__0^6)*wit__1^2 + (t_0*wit__0^4*wit__1^4 + (3*wit_^2*wit__0^4 - wit__0^6)*wit__1^2)*wit__11^2)*wit__13^2)*wit__9^6 - 1/4*(4*(wit__1^2*wit__11^2 + (wit_^2 + wit__0^2)*wit__1^2)*wit__12^4*wit__2^2 - (4*wit_^2*wit__11^4 + wit__11^6 - 4*(h_0*wit_^2 + h_0*wit__0^2)*wit__1^2 + 4*(2*wit_^4 - h_0*wit__1^2)*wit__11^2)*wit__12^4 + 6*(t_0*wit__0^4*wit__1^6 + wit__0^4*wit__1^4*wit__11^2 + (2*wit_^2*wit__0^4 - wit__0^6)*wit__1^4)*wit__13^4 - (4*wit_^2*wit__11^4 + wit__11^6 - 4*h_0*wit__1^2*wit__11^2 - 4*(h_0*wit_^2 + h_0*wit__0^2)*wit__1^2 - 4*(wit__1^2*wit__11^2 + (wit_^2 + wit__0^2)*wit__1^2)*wit__2^2)*wit__3^4 + 2*(4*(wit__1^2*wit__11^2 + (wit_^2 + wit__0^2)*wit__1^2)*wit__12^2*wit__2^2 - (4*wit_^2*wit__11^4 + wit__11^6 - 4*(h_0*wit_^2 + h_0*wit__0^2)*wit__1^2 + 4*(wit_^4 - h_0*wit__1^2)*wit__11^2)*wit__12^2)*wit__3^2 + 2*(2*wit__0^2*wit__12^2*wit__3^2 - (wit_^2 - wit__0^2)*wit__12^4 + (wit_^2 + wit__0^2)*wit__3^4)*wit__6^2 + 4*(wit__1^2*wit__12^4*wit__2^2 + (wit_^2*wit__11^2 + h_0*wit__1^2)*wit__12^4 - (wit_^2*wit__11^2 - wit__1^2*wit__2^2 - h_0*wit__1^2)*wit__3^4 + 2*(wit__1^2*wit__12^2*wit__2^2 + h_0*wit__1^2*wit__12^2)*wit__3^2)*wit__7^2)*wit__9^4 + 1/2*(wit__12^8 - 4*wit__12^4*wit__3^4 - 4*wit__12^2*wit__3^6 - wit__3^8)*wit__7^2 - 1/4*(3*wit__0^4*wit__1^6*wit__13^6 - 8*wit_^2*wit__0^2*wit__1^2*wit__12^2*wit__4^2 - (8*wit_^4 - 10*wit_^2*wit__0^2 + wit__0^4 - wit__11^4 + 4*h_0*wit__1^2 - (4*wit_^2 + wit__0^2)*wit__11^2)*wit__12^6 - (3*wit__1^2*wit__11^4 - 4*h_0*wit__1^4 + 2*(4*wit_^2 - wit__0^2)*wit__1^2*wit__11^2 + (4*wit_^4 - 4*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__12^4*wit__13^2 - (3*wit__0^2*wit__1^4*wit__11^2 + (4*wit_^2*wit__0^2 - wit__0^4)*wit__1^4)*wit__12^2*wit__13^4 - 2*(2*wit_^4 + wit_^2*wit__0^2 + (4*wit_^2 - wit__0^2)*wit__11^2)*wit__3^6 - ((8*wit_^4 + 2*wit_^2*wit__0^2 + 2*wit__0^4 - wit__11^4 + 4*h_0*wit__1^2 + 6*(2*wit_^2 - wit__0^2)*wit__11^2)*wit__12^2 + (8*wit_^2*wit__1^2*wit__11^2 + 3*wit__1^2*wit__11^4 - 4*h_0*wit__1^4 + 2*(2*wit_^4 + 4*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__13^2 - 4*(wit__1^4*wit__13^2 - wit__1^2*wit__12^2)*wit__2^2)*wit__3^4 + 4*(wit__1^4*wit__12^4*wit__13^2 - wit__1^2*wit__12^6)*wit__2^2 - ((12*wit_^4 - 6*wit_^2*wit__0^2 + 4*wit__0^4 - 5*wit__0^2*wit__11^2 - 2*wit__11^4 + 8*h_0*wit__1^2)*wit__12^4 + 2*(3*wit__1^2*wit__11^4 - 4*h_0*wit__1^4 + (8*wit_^2 - wit__0^2)*wit__1^2*wit__11^2 + 2*(2*wit_^4 + wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__12^2*wit__13^2 + (4*wit_^2*wit__0^2*wit__1^4 + 3*wit__0^2*wit__1^4*wit__11^2)*wit__13^4 - 8*(wit__1^4*wit__12^2*wit__13^2 - wit__1^2*wit__12^4)*wit__2^2)*wit__3^2 - 2*(wit__12^6 + 3*wit__12^4*wit__3^2 + 3*wit__12^2*wit__3^4 + wit__3^6)*wit__6^2 + 4*(wit_^2*wit__1^2*wit__12^4*wit__13^2 + wit_^2*wit__12^6 - wit_^2*wit__12^4*wit__3^2 - wit_^2*wit__3^6 - (wit_^2*wit__1^2*wit__13^2 + 3*wit_^2*wit__12^2)*wit__3^4)*wit__7^2)*wit__9^2"),(5,"-1/2*(2*wit_^4 - 2*wit_^2*wit__0^2 + h_0*wit__1^2)*wit__12^6 - 1/4*(wit__0^2*wit__1^2*wit__11^2 + wit__1^2*wit__11^4 - 4*h_0*wit__1^4 - (2*wit_^2*wit__0^2 + wit__0^4)*wit__1^2)*wit__12^4*wit__13^2 + 1/4*(3*wit__1^4*wit__11^4 - 4*h_0*wit__1^6 + 2*(4*wit_^2 - wit__0^2)*wit__1^4*wit__11^2 + (4*wit_^4 - wit__0^4)*wit__1^4)*wit__12^2*wit__13^4 + 1/4*(3*wit__0^2*wit__1^6*wit__11^2 + (4*wit_^2*wit__0^2 - wit__0^4)*wit__1^6)*wit__13^6 - 1/4*(4*wit_^4 - wit__0^4 - 2*wit__1^2*wit__2^2 - 2*h_0*wit__1^2 + 2*(6*wit_^2 - wit__0^2)*wit__11^2)*wit__3^6 + 1/4*(4*wit_^4*wit__0^6 - 4*wit_^2*wit__0^8 + wit__0^10 + 4*wit_^2*wit__0^2*wit__11^6 + wit__0^2*wit__11^8 + 4*wit__0^2*wit__1^4*wit__2^4 + 4*h_0^2*wit__0^2*wit__1^4 + 2*(2*wit_^4*wit__0^2 + 2*wit_^2*wit__0^4 - wit__0^6 - 2*h_0*wit__0^2*wit__1^2)*wit__11^4 - 4*(2*h_0*wit_^2*wit__0^4 - h_0*wit__0^6)*wit__1^2 + 4*(2*wit_^4*wit__0^4 - wit_^2*wit__0^6 - 2*h_0*wit_^2*wit__0^2*wit__1^2)*wit__11^2 - 4*(2*wit_^2*wit__0^2*wit__1^2*wit__11^2 + wit__0^2*wit__1^2*wit__11^4 - 2*h_0*wit__0^2*wit__1^4 + (2*wit_^2*wit__0^4 - wit__0^6)*wit__1^2)*wit__2^2)*wit__9^6 - 1/4*((4*wit_^4 + 4*wit_^2*wit__0^2 - wit__0^4 - wit__11^4 - 2*h_0*wit__1^2 + 4*(2*wit_^2 - wit__0^2)*wit__11^2)*wit__12^2 + (3*wit__1^2*wit__11^4 - 4*h_0*wit__1^4 + 2*(2*wit_^2 + wit__0^2)*wit__1^2*wit__11^2 + 2*(3*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__13^2 - 2*(2*wit__1^4*wit__13^2 + wit__1^2*wit__12^2)*wit__2^2)*wit__3^4 - 2*(wit_^2*wit__0^2*wit__1^2 + wit_^2*wit__1^2*wit__11^2)*wit__12^2*wit__4^2 + 1/4*(4*wit__1^4*wit__12^2*wit__2^4 + (4*wit_^4*wit__0^4 - 4*wit_^2*wit__0^6 + wit__0^8 + 4*wit_^2*wit__11^6 + wit__11^8 + 4*h_0^2*wit__1^4 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4 - 2*h_0*wit__1^2)*wit__11^4 - 4*(2*h_0*wit_^2*wit__0^2 - h_0*wit__0^4)*wit__1^2 + 4*(2*wit_^4*wit__0^2 - wit_^2*wit__0^4 - 2*h_0*wit_^2*wit__1^2)*wit__11^2)*wit__12^2 + 4*(3*wit_^2*wit__0^2*wit__1^2*wit__11^4 + wit__0^2*wit__1^2*wit__11^6 - 2*h_0*wit_^2*wit__0^2*wit__1^4 + (2*wit_^4*wit__0^4 - wit_^2*wit__0^6)*wit__1^2 - (2*h_0*wit__0^2*wit__1^4 - (2*wit_^4*wit__0^2 + 2*wit_^2*wit__0^4 - wit__0^6)*wit__1^2)*wit__11^2)*wit__13^2 - 4*((2*wit_^2*wit__1^2*wit__11^2 + wit__1^2*wit__11^4 - 2*h_0*wit__1^4 + (2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__12^2 + 2*(wit_^2*wit__0^2*wit__1^4 + wit__0^2*wit__1^4*wit__11^2)*wit__13^2)*wit__2^2 + (4*wit_^4*wit__0^4 - 4*wit_^2*wit__0^6 + wit__0^8 + 4*wit_^2*wit__11^6 + wit__11^8 + 4*wit__1^4*wit__2^4 + 4*h_0^2*wit__1^4 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4 - 2*h_0*wit__1^2)*wit__11^4 - 4*(2*h_0*wit_^2*wit__0^2 - h_0*wit__0^4)*wit__1^2 + 4*(2*wit_^4*wit__0^2 - wit_^2*wit__0^4 - 2*h_0*wit_^2*wit__1^2)*wit__11^2 - 4*(2*wit_^2*wit__1^2*wit__11^2 + wit__1^2*wit__11^4 - 2*h_0*wit__1^4 + (2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__2^2)*wit__3^2)*wit__9^4 - 1/2*(2*wit__1^6*wit__12^2*wit__13^4 - 2*wit__1^4*wit__12^4*wit__13^2 + wit__1^2*wit__12^6)*wit__2^2 - 1/4*((4*wit_^4 - 4*wit_^2*wit__0^2 + wit__0^4 - wit__0^2*wit__11^2 - wit__11^4 + 2*h_0*wit__1^2)*wit__12^4 + 2*(2*wit__1^2*wit__11^4 - 4*h_0*wit__1^4 + (2*wit_^2 + wit__0^2)*wit__1^2*wit__11^2 - 2*(wit_^2*wit__0^2 + wit__0^4)*wit__1^2)*wit__12^2*wit__13^2 - (3*wit__1^4*wit__11^4 - 4*h_0*wit__1^6 + (8*wit_^2 - 3*wit__0^2)*wit__1^4*wit__11^2 + 2*(2*wit_^4 - wit__0^4)*wit__1^4)*wit__13^4 + 2*(2*wit__1^6*wit__13^4 - 4*wit__1^4*wit__12^2*wit__13^2 + wit__1^2*wit__12^4)*wit__2^2)*wit__3^2 + (2*wit_^2*wit__0^2*wit__1^4*wit__12^2*wit__13^2 - wit_^2*wit__0^2*wit__1^2*wit__12^2*wit__3^2 - wit_^2*wit__1^2*wit__11^2*wit__3^4 + (wit_^2*wit__0^2*wit__1^2 + wit_^2*wit__1^2*wit__11^2)*wit__12^4)*wit__5^2 - 1/2*(2*wit__12^4*wit__3^2 + 4*wit__12^2*wit__3^4 + 2*wit__3^6 + (wit__1^2*wit__12^4 + 2*wit__1^2*wit__12^2*wit__3^2 + wit__1^2*wit__3^4)*wit__5^2)*wit__6^2 + (2*wit_^2*wit__1^2*wit__12^4*wit__13^2 + 2*wit_^2*wit__1^2*wit__12^2*wit__13^2*wit__3^2 + wit_^2*wit__12^6 - wit_^2*wit__12^2*wit__3^4 + (2*wit_^2*wit__1^4*wit__12^2*wit__13^2 + wit_^2*wit__1^2*wit__12^4 - wit_^2*wit__1^2*wit__12^2*wit__3^2)*wit__5^2)*wit__7^2 + 1/4*((4*wit_^4*wit__0^2 - 6*wit_^2*wit__0^4 + wit__0^6 - wit__11^6 + 2*h_0*wit__0^2*wit__1^2 - 2*(2*wit_^2 + wit__0^2)*wit__11^4 - 2*(3*wit_^2*wit__0^2 - 2*h_0*wit__1^2)*wit__11^2)*wit__12^4 + 4*(3*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 - 2*h_0*wit_^2*wit__1^4 + 2*(wit_^4*wit__0^2 - wit_^2*wit__0^4)*wit__1^2 - (2*h_0*wit__1^4 - (2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__11^2)*wit__12^2*wit__13^2 + 2*(6*wit_^2*wit__0^2*wit__1^4*wit__11^2 + 3*wit__0^2*wit__1^4*wit__11^4 - 2*h_0*wit__0^2*wit__1^6 + (2*wit_^4*wit__0^2 + 2*wit_^2*wit__0^4 - wit__0^6)*wit__1^4)*wit__13^4 + (4*wit_^4*wit__0^2 - 4*wit_^2*wit__0^4 + wit__0^6 - wit__11^6 + 2*h_0*wit__0^2*wit__1^2 + 2*(2*wit_^2 - wit__0^2)*wit__11^4 + 2*(4*wit_^4 - wit_^2*wit__0^2 + 2*h_0*wit__1^2)*wit__11^2 + 2*(wit__0^2*wit__1^2 + 2*wit__1^2*wit__11^2)*wit__2^2)*wit__3^4 - 2*(2*wit__0^2*wit__1^6*wit__13^4 - (wit__0^2*wit__1^2 + 2*wit__1^2*wit__11^2)*wit__12^4 + 4*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__12^2*wit__13^2)*wit__2^2 + 2*((4*wit_^4*wit__0^2 - 3*wit_^2*wit__0^4 + wit__0^6 - 2*wit__0^2*wit__11^4 - wit__11^6 + 2*h_0*wit__0^2*wit__1^2 + 4*(wit_^4 - wit_^2*wit__0^2 + h_0*wit__1^2)*wit__11^2)*wit__12^2 + 2*(3*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 - 2*h_0*wit_^2*wit__1^4 + (2*wit_^4*wit__0^2 - wit_^2*wit__0^4)*wit__1^2 - (2*h_0*wit__1^4 - (2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^2)*wit__11^2)*wit__13^2 + 2*((wit__0^2*wit__1^2 + 2*wit__1^2*wit__11^2)*wit__12^2 - 2*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__13^2)*wit__2^2)*wit__3^2 + 2*(2*wit_^2*wit__1^2*wit__12^2*wit__13^2 + (wit_^2 + wit__11^2)*wit__12^4 + (wit_^2 + 2*wit__11^2)*wit__12^2*wit__3^2 + (2*wit_^2 + wit__11^2)*wit__3^4)*wit__6^2 + 2*(2*wit_^2*wit__11^2*wit__12^4 - 2*wit_^2*wit__11^2*wit__3^4 - (wit__12^4 + 2*wit__12^2*wit__3^2 + wit__3^4)*wit__6^2)*wit__7^2)*wit__9^2"),(6,"-1/4*wit__0^4*wit__1^8*wit__13^4*wit__5^2 + 1/4*wit__0^4*wit__1^8*wit__13^2*wit__5^4 - 1/4*wit__0^4*wit__1^8*wit__5^6 - 2*wit_^2*wit__0^2*wit__1^4*wit__12^2*wit__4^2 + 1/4*(wit__0^2*wit__1^2*wit__11^2 + wit__1^2*wit__11^4 - 4*h_0*wit__1^4 - (8*wit_^4 - 6*wit_^2*wit__0^2 + wit__0^4)*wit__1^2)*wit__12^6 - 1/4*(3*wit__1^4*wit__11^4 - 4*h_0*wit__1^6 + 2*(4*wit_^2 - wit__0^2)*wit__1^4*wit__11^2 + (4*wit_^4 - wit__0^4)*wit__1^4)*wit__12^4*wit__13^2 - 1/4*(3*wit__0^2*wit__1^6*wit__11^2 + (4*wit_^2*wit__0^2 - wit__0^4)*wit__1^6)*wit__12^2*wit__13^4 - 1/2*((2*wit_^2 - wit__0^2)*wit__1^2*wit__11^2 + (2*wit_^4 - wit_^2*wit__0^2)*wit__1^2)*wit__3^6 + (t_0^3*wit__0^4*wit__1^8 - 2*t_0*wit__0^4*wit__1^6*wit__2^2 + 2*t_0*wit__0^4*wit__1^4*wit__7^4 - 2*h_0*t_0*wit__0^4*wit__1^6 + (3*t_0^2*wit__0^4*wit__1^6 - 2*wit__0^4*wit__1^4*wit__2^2 - 2*h_0*wit__0^4*wit__1^4)*wit__7^2)*wit__9^6 + 1/4*((wit__1^2*wit__11^4 - 4*h_0*wit__1^4 - 2*(4*wit_^2 - 3*wit__0^2)*wit__1^2*wit__11^2 - 2*(4*wit_^4 - 5*wit_^2*wit__0^2 + wit__0^4)*wit__1^2)*wit__12^2 - (8*wit_^2*wit__1^4*wit__11^2 + 3*wit__1^4*wit__11^4 - 4*h_0*wit__1^6 + 2*(2*wit_^4 + 2*wit_^2*wit__0^2 - wit__0^4)*wit__1^4)*wit__13^2 + 4*(wit__1^6*wit__13^2 - wit__1^4*wit__12^2)*wit__2^2)*wit__3^4 - 1/2*(2*wit__0^4*wit__1^6*wit__13^2*wit__2^2 - 2*(wit__0^4*wit__1^4*wit__13^2 - wit__0^4*wit__1^4*wit__5^2)*wit__7^4 - (3*t_0^2*wit__0^4*wit__1^8 - 2*h_0*wit__0^4*wit__1^6)*wit__13^2 + (3*t_0^2*wit__0^4*wit__1^8 - 2*wit__0^4*wit__1^6*wit__2^2 - 2*h_0*wit__0^4*wit__1^6)*wit__5^2 - 6*(t_0*wit__0^4*wit__1^6*wit__13^2 - t_0*wit__0^4*wit__1^6*wit__5^2)*wit__7^2)*wit__9^4 + (wit__1^6*wit__12^4*wit__13^2 - wit__1^4*wit__12^6)*wit__2^2 + 1/4*((2*wit__1^2*wit__11^4 - 8*h_0*wit__1^4 - (4*wit_^2 - 5*wit__0^2)*wit__1^2*wit__11^2 - 2*(6*wit_^4 - 5*wit_^2*wit__0^2 + 2*wit__0^4)*wit__1^2)*wit__12^4 - 2*(3*wit__1^4*wit__11^4 - 4*h_0*wit__1^6 + (8*wit_^2 - wit__0^2)*wit__1^4*wit__11^2 + 2*(2*wit_^4 + wit_^2*wit__0^2 - wit__0^4)*wit__1^4)*wit__12^2*wit__13^2 - (4*wit_^2*wit__0^2*wit__1^6 + 3*wit__0^2*wit__1^6*wit__11^2)*wit__13^4 + 8*(wit__1^6*wit__12^2*wit__13^2 - wit__1^4*wit__12^4)*wit__2^2)*wit__3^2 + 1/4*(4*t_0*wit__0^4*wit__1^8*wit__13^4 - 4*t_0*wit__0^4*wit__1^8*wit__13^2*wit__5^2 + 4*t_0*wit__0^4*wit__1^8*wit__5^4 + 4*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__12^4*wit__2^2 - (4*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 - 4*h_0*wit_^2*wit__1^4 + 4*(2*wit_^4*wit__1^2 - h_0*wit__1^4)*wit__11^2)*wit__12^4 - (4*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 - 4*h_0*wit_^2*wit__1^4 - 4*h_0*wit__1^4*wit__11^2 - 4*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__2^2)*wit__3^4 + 2*(4*(wit_^2*wit__1^4 + wit__1^4*wit__11^2)*wit__12^2*wit__2^2 - (4*wit_^2*wit__1^2*wit__11^4 + wit__1^2*wit__11^6 - 4*h_0*wit_^2*wit__1^4 + 4*(wit_^4*wit__1^2 - h_0*wit__1^4)*wit__11^2)*wit__12^2)*wit__3^2 - 2*(wit_^2*wit__1^2*wit__12^4 - wit_^2*wit__1^2*wit__3^4)*wit__6^2 + 4*(wit__0^4*wit__1^6*wit__13^4 - wit__0^4*wit__1^6*wit__13^2*wit__5^2 + wit__0^4*wit__1^6*wit__5^4)*wit__7^2)*wit__9^2"),(7,"-2*wit_^2*wit__1^2*wit__12^6*wit__13^2 - 2*wit_^2*wit__1^2*wit__12^4*wit__13^2*wit__3^2 - 1/4*(4*wit_^2 - wit__0^2 - 2*wit__11^2)*wit__12^8 + (wit_^2 - 2*wit__0^2 - 2*wit__11^2)*wit__12^4*wit__3^4 - 2*(wit__0^2 + wit__11^2)*wit__12^2*wit__3^6 - 1/2*(wit__0^2 + wit__11^2)*wit__3^8 - (t_0^3*wit__0^4*wit__1^6 + t_0^2*wit__0^4*wit__1^4*wit__7^2 - 2*t_0*wit__0^4*wit__1^4*wit__2^2 - (t_0^2*wit__0^6 - (t_0^2*wit_^2 - 2*h_0*t_0)*wit__0^4)*wit__1^4)*wit__9^8 - (2*t_0*wit__0^4*wit__1^4*wit__13^2*wit__7^2 - 2*wit__0^4*wit__1^4*wit__13^2*wit__2^2 + (3*t_0^2*wit__0^4*wit__1^6 - 2*(t_0*wit__0^6 - (t_0*wit_^2 - h_0)*wit__0^4)*wit__1^4)*wit__13^2)*wit__9^6 - 1/2*(2*wit__0^4*wit__1^4*wit__13^4*wit__7^2 + 2*(wit__0^2*wit__1^2 + wit__1^2*wit__11^2)*wit__12^4*wit__2^2 + 2*(wit_^2*wit__11^4 + h_0*wit__0^2*wit__1^2 + h_0*wit__1^2*wit__11^2)*wit__12^4 + 2*(3*t_0*wit__0^4*wit__1^6 + (wit_^2*wit__0^4 - wit__0^6)*wit__1^4)*wit__13^4 - 2*(wit_^2*wit__11^4 - h_0*wit__0^2*wit__1^2 - h_0*wit__1^2*wit__11^2 - (wit__0^2*wit__1^2 + wit__1^2*wit__11^2)*wit__2^2)*wit__3^4 + 4*((wit__0^2*wit__1^2 + wit__1^2*wit__11^2)*wit__12^2*wit__2^2 + (h_0*wit__0^2*wit__1^2 + h_0*wit__1^2*wit__11^2)*wit__12^2)*wit__3^2 + (wit__0^2*wit__12^4 + 2*wit__0^2*wit__12^2*wit__3^2 + wit__0^2*wit__3^4)*wit__6^2)*wit__9^4 - 1/2*(2*wit__0^4*wit__1^6*wit__13^6 + 4*wit_^2*wit__11^2*wit__12^6 - 2*(wit_^2*wit__0^2*wit__1^2 - wit_^2*wit__1^2*wit__11^2)*wit__12^4*wit__13^2 - 2*(wit_^2*wit__0^2 + wit_^2*wit__11^2)*wit__3^6 - 2*(2*(wit_^2*wit__0^2 + 2*wit_^2*wit__11^2)*wit__12^2 + (wit_^2*wit__0^2*wit__1^2 + wit_^2*wit__1^2*wit__11^2)*wit__13^2)*wit__3^4 - 2*(2*wit_^2*wit__0^2*wit__1^2*wit__12^2*wit__13^2 + (wit_^2*wit__0^2 + wit_^2*wit__11^2)*wit__12^4)*wit__3^2 - (wit__12^6 + 2*wit__12^4*wit__3^2 + wit__12^2*wit__3^4)*wit__6^2)*wit__9^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(1),List((0,"w_0","-1","1"),(0,"abs_1","rv_0*t_0 - r_0","1"),(3,"r_0","rv_0*t_0 - wit__1^2","1"),(10,"rv_0","wit__9^2","1"),(0,"min_1","0","1"),(0,"dhd_0","dhf_0 + max_1","-1"),(7,"dhf_0","-wit__8^2 - max_1","1"),(0,"max_1","wit__0^2","-1"),(0,"a_0","wit__2^2","1"),(0,"hp_0","wit__3^2","1"),(2,"rp_0","wit__1^2 + wit__6^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__1*wit__10"),("1","wit__1*wit__13")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"wit__1^2"),(4,"0"),(5,"0"),(6,"-wit__1^2"),(7,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(1),List((0,"w_0","1","1"),(0,"abs_1","rv_0*t_0 - r_0","1"),(3,"r_0","rv_0*t_0 - wit__1^2","1"),(10,"rv_0","wit__9^2","1"),(0,"min_1","0","1"),(0,"dhd_0","dhf_0 - max_1","1"),(7,"dhf_0","wit__8^2 + max_1","-1"),(0,"max_1","wit__0^2","-1"),(0,"a_0","wit__2^2","1"),(0,"hp_0","wit__3^2","1"),(2,"rp_0","wit__1^2 + wit__6^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__1*wit__10"),("1","wit__1*wit__13")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"wit__1^2"),(4,"0"),(5,"0"),(6,"-wit__1^2"),(7,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(2,5),List((0,"w_0","-1","1"),(0,"abs_1","-rv_0*t_0 + r_0","1"),(8,"r_0","rv_0*t_0 - wit__6^2 + rp_0","-1"),(10,"rv_0","wit__9^2","1"),(0,"dhd_0","-min_1","1"),(0,"min_1","-dhf_0 - max_1","1"),(1,"dhf_0","wit__1^2 - max_1","1"),(0,"max_1","wit__0^2","-1"),(0,"a_0","wit__2^2","1"),(0,"hp_0","wit__3^2","1"),(3,"rp_0","wit__6^2 + wit__8^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1/2","wit__2*wit__9^2*wit__5*wit__7"),("5/12","wit__2*wit__9^2*wit__5*wit__11*wit_"),("1/3","wit__2*wit__4*wit__5*wit__12"),("1/4","wit__2*wit__9*wit__5*wit__11*wit__12"),("1/12","-wit__2^3*wit__5*wit__13^2 + wit__2*wit__5*wit__12^2"),("1/12","wit__2^2*wit__9*wit__5*wit__11*wit__13"),("1/12","wit__2*wit__9*wit__5*wit__12*wit_"),("7/12","wit__2^2*wit__4*wit__6*wit__5"),("1/12","wit__2^2*wit__6*wit__5*wit__12")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"wit__13^6*wit__2^6 + wit__12^2*wit__13^4*wit__2^4 + wit__13^4*wit__2^4*wit__4^2 + 1/4*(4*wit_^2*wit__0^4 - 2*wit__0^6 + 4*wit__11^6 + 4*(wit_^2 + wit__0^2)*wit__11^4 - 4*(wit__0^2 + wit__11^2)*wit__2^2*wit__3^2 + 2*(4*wit_^2*wit__0^2 - wit__0^4)*wit__11^2 + 4*(h_0*wit__0^2 + h_0*wit__11^2)*wit__2^2 + (t_0*wit__2^2 + wit__0^2 + wit__11^2)*wit__7^2)*wit__9^6 + 1/12*wit__12^2*wit__2^2*wit__5^2 - 1/2*(4*wit__13^2*wit__2^4*wit__3^2 - 4*h_0*wit__13^2*wit__2^4 + wit__2^2*wit__6^2*wit__7^2 - 2*(2*wit_^2*wit__0^2 - wit__0^4 + 3*wit__11^4 + 2*(wit_^2 + wit__0^2)*wit__11^2)*wit__13^2*wit__2^2 - 2*(wit__0^2*wit__11^2 + wit__11^4)*wit__12^2 - 2*(wit__0^2*wit__11^2 + wit__11^4)*wit__4^2)*wit__9^4 + 1/2*(2*wit__12^2*wit__13^2*wit__2^4 + 2*wit__13^2*wit__2^4*wit__4^2 + wit__12^4*wit__2^2)*wit__6^2 + 1/12*(5*wit__11^2*wit__2^2*wit__5^2 + 3*(wit__0^2 + wit__11^2)*wit__12^4 - 6*(t_0*wit__12^2*wit__13^2 - 2*(wit_^2 + wit__0^2 + 3*wit__11^2)*wit__13^4)*wit__2^4 - 3*(t_0*wit__12^4 - 6*(wit__0^2 + wit__11^2)*wit__12^2*wit__13^2)*wit__2^2 - 6*(t_0*wit__13^2*wit__2^4 - 3*(wit__0^2 + wit__11^2)*wit__13^2*wit__2^2 - 2*(wit__0^2 + wit__11^2)*wit__12^2)*wit__4^2)*wit__9^2"),(1,"wit__12^2*wit__13^2*wit__2^4 + wit__13^2*wit__2^4*wit__4^2 - 1/2*wit__2^2*wit__7^2*wit__9^4 + 1/2*wit__12^4*wit__2^2"),(2,"-1/4*(wit__0^4 + 2*wit__2^2*wit__3^2 - 2*h_0*wit__2^2)*wit__9^8 + 1/2*(wit__0^2*wit__12^2 + wit__0^2*wit__4^2)*wit__9^6 - 1/4*(wit__12^4 + 2*wit__12^2*wit__4^2)*wit__9^4"),(3,"0"),(4,"-1/4*(t_0^3*wit__2^6 + 4*wit_^4*wit__0^2 - 4*wit_^2*wit__0^4 + 3*wit__0^6 - 2*wit__0^4*wit__1^2 + (4*wit_^2 - wit__0^2)*wit__11^4 + (4*t_0^2*wit_^2 - 3*t_0^2*wit__0^2 + t_0^2*wit__11^2 + 4*h_0*t_0)*wit__2^4 + (4*wit_^4 + wit__0^4)*wit__11^2 + (4*t_0*wit_^4 + t_0*wit__0^4 + t_0*wit__11^4 + 8*h_0*wit_^2 - 4*(t_0*wit_^2 + 2*h_0)*wit__0^2 + 4*h_0*wit__1^2 + 2*(2*t_0*wit_^2 - t_0*wit__0^2)*wit__11^2)*wit__2^2 - 4*(t_0*wit__2^4 + (2*wit_^2 - 2*wit__0^2 + wit__1^2)*wit__2^2)*wit__3^2 + (wit__0^2 - wit__1^2)*wit__7^2)*wit__9^8 - 1/4*(4*t_0^2*wit__13^2*wit__2^6 - 4*wit__13^2*wit__2^4*wit__3^2 + 4*(3*t_0*wit_^2 - 2*t_0*wit__0^2 + t_0*wit__11^2 + h_0)*wit__13^2*wit__2^4 + (8*wit_^4 - 4*wit_^2*wit__0^2 + 2*wit__0^4 + wit__11^4 + 4*(3*wit_^2 - wit__0^2)*wit__11^2)*wit__13^2*wit__2^2 - (4*wit__0^4 - 4*wit__0^2*wit__1^2 + 4*wit__0^2*wit__11^2 + 3*wit__11^4)*wit__12^2 - 2*(wit__0^4 - 2*wit__0^2*wit__1^2 + 2*wit__0^2*wit__11^2 + 2*wit__11^4 - 2*wit__2^2*wit__3^2 + 2*h_0*wit__2^2)*wit__4^2 - (wit__12^2 + wit__4^2)*wit__7^2)*wit__9^6 - 1/4*(6*t_0*wit__13^4*wit__2^6 + 3*(4*wit_^2 - 2*wit__0^2 + wit__11^2)*wit__13^4*wit__2^4 - 2*(3*wit__0^2 - wit__1^2 + 3*wit__11^2)*wit__12^2*wit__13^2*wit__2^2 + (wit__0^2 - wit__1^2 - wit__11^2)*wit__12^4 - 4*(wit__0^2 + wit__11^2)*wit__4^4 - 2*((3*wit__0^2 - wit__1^2 + 4*wit__11^2)*wit__13^2*wit__2^2 + 2*(wit__1^2 + wit__11^2)*wit__12^2)*wit__4^2)*wit__9^4 - 1/12*(9*wit__13^6*wit__2^6 - 9*wit__12^2*wit__13^4*wit__2^4 + 3*wit__12^4*wit__13^2*wit__2^2 - 6*wit__13^2*wit__2^2*wit__4^4 - 3*(4*wit__13^4*wit__2^4 + wit__12^4)*wit__4^2 - (wit__13^2*wit__2^4 + 8*wit__12^2*wit__2^2 + 5*wit__2^2*wit__4^2)*wit__5^2)*wit__9^2"),(5,"3/4*wit__13^6*wit__2^6 + 1/4*wit__12^2*wit__13^4*wit__2^4 + 1/4*wit__12^4*wit__13^2*wit__2^2 + 1/2*wit__12^2*wit__13^2*wit__2^2*wit__4^2 + 1/4*(4*wit_^2*wit__0^4 - 2*wit__0^6 + 8*wit_^2*wit__0^2*wit__11^2 + 3*wit__11^6 - 4*wit__0^2*wit__2^2*wit__3^2 + 4*(wit_^2 + wit__0^2)*wit__11^4 + 4*h_0*wit__0^2*wit__2^2 + (2*wit__0^2 + wit__11^2)*wit__7^2)*wit__9^6 - 1/4*(4*wit__13^2*wit__2^4*wit__3^2 - 4*h_0*wit__13^2*wit__2^4 - wit__11^4*wit__12^2 - (8*wit_^2*wit__0^2 - 2*wit__0^4 + 9*wit__11^4 + 8*(wit_^2 + wit__0^2)*wit__11^2)*wit__13^2*wit__2^2 - 2*(wit__0^4 + 2*wit__2^2*wit__3^2 - 2*h_0*wit__2^2)*wit__4^2 + (wit__2^2*wit__6^2 + wit__12^2 + 2*wit__4^2)*wit__7^2)*wit__9^4 - 1/12*(wit__13^2*wit__2^4 + 7*wit__12^2*wit__2^2 + 5*wit__2^2*wit__4^2)*wit__5^2 + 1/4*(2*wit__12^2*wit__13^2*wit__2^4 + 2*wit__13^2*wit__2^4*wit__4^2 + wit__12^4*wit__2^2)*wit__6^2 + 1/12*(3*(4*wit_^2 + 4*wit__0^2 + 9*wit__11^2)*wit__13^4*wit__2^4 - 6*wit__11^2*wit__13^2*wit__2^2*wit__4^2 + 5*wit__11^2*wit__2^2*wit__5^2 - 12*(wit__0^2 + wit__11^2)*wit__4^4)*wit__9^2"),(6,"-1/4*wit__13^4*wit__2^8*wit__6^2 + 1/4*wit__13^2*wit__2^8*wit__6^4 - 1/4*wit__2^8*wit__6^6 - 3/4*wit__12^2*wit__13^4*wit__2^6 - 1/4*wit__12^4*wit__13^2*wit__2^4 - wit__13^2*wit__2^4*wit__4^4 - 1/4*wit__12^6*wit__2^2 + 1/4*(4*t_0^3*wit__2^8 - 4*(3*t_0^2*wit__1^2 - 2*h_0*t_0)*wit__2^6 + 4*(2*t_0*wit__1^4 - 2*h_0*wit__1^2 - h_0*wit__11^2)*wit__2^4 + (2*wit__0^4*wit__11^2 - wit__11^6)*wit__2^2 - 4*(2*t_0*wit__2^6 - (2*wit__1^2 + wit__11^2)*wit__2^4)*wit__3^2)*wit__9^6 + 1/4*(6*t_0^2*wit__13^2*wit__2^8 - 12*t_0*wit__1^2*wit__13^2*wit__2^6 + (2*wit__0^4 + 4*wit__1^4 - 3*wit__11^4)*wit__13^2*wit__2^4 - (4*wit__0^2*wit__11^2 + 3*wit__11^4)*wit__12^2*wit__2^2 + 2*(2*wit__2^4*wit__3^2 - 2*h_0*wit__2^4 + (wit__0^4 - 2*wit__0^2*wit__11^2 - 2*wit__11^4)*wit__2^2)*wit__4^2 - 2*(3*t_0^2*wit__2^8 + 2*wit__1^4*wit__2^4 - 2*wit__2^6*wit__3^2 - 2*(3*t_0*wit__1^2 - h_0)*wit__2^6)*wit__6^2)*wit__9^4 - 1/2*(2*wit__13^4*wit__2^6 + 2*wit__12^2*wit__13^2*wit__2^4 + wit__12^4*wit__2^2)*wit__4^2 - 1/12*(wit__13^2*wit__2^6 + 9*wit__12^2*wit__2^4 + 12*wit__2^4*wit__4^2)*wit__5^2 + 1/4*(4*t_0*wit__13^4*wit__2^8 - (4*wit__1^2 + 3*wit__11^2)*wit__13^4*wit__2^6 - 2*(2*wit__0^2 + 3*wit__11^2)*wit__12^2*wit__13^2*wit__2^4 - wit__11^2*wit__12^4*wit__2^2 - 4*(wit__0^2 + wit__11^2)*wit__2^2*wit__4^4 + 4*(t_0*wit__2^8 - wit__1^2*wit__2^6)*wit__6^4 - 4*((wit__0^2 + 2*wit__11^2)*wit__13^2*wit__2^4 + (wit__0^2 + wit__11^2)*wit__12^2*wit__2^2)*wit__4^2 - 4*(t_0*wit__13^2*wit__2^8 - wit__1^2*wit__13^2*wit__2^6)*wit__6^2)*wit__9^2"),(7,"-(t_0^3*wit__2^6 - wit_^2*wit__0^4 + wit__0^6 - wit__11^6 - (wit_^2 + wit__0^2)*wit__11^4 + (t_0^2*wit_^2 - t_0^2*wit__0^2 - t_0^2*wit__1^2 + 2*h_0*t_0)*wit__2^4 - (2*wit_^2*wit__0^2 - wit__0^4)*wit__11^2 - 2*(h_0*wit__0^2 + h_0*wit__11^2)*wit__2^2 - 2*(t_0*wit__2^4 - (wit__0^2 + wit__11^2)*wit__2^2)*wit__3^2)*wit__9^8 - (3*t_0^2*wit__13^2*wit__2^6 + 2*(t_0*wit_^2 - t_0*wit__0^2 - t_0*wit__1^2)*wit__13^2*wit__2^4 - (2*wit_^2*wit__0^2 - wit__0^4 + 3*wit__11^4 + 2*(wit_^2 + wit__0^2)*wit__11^2)*wit__13^2*wit__2^2 - (wit__0^4 + 2*wit__0^2*wit__11^2 + wit__11^4)*wit__12^2 - (wit__0^4 + 2*wit__0^2*wit__11^2 + wit__11^4)*wit__4^2)*wit__9^6 - 1/12*(36*t_0*wit__13^4*wit__2^6 - 12*(2*wit__0^2 + wit__1^2 + 3*wit__11^2)*wit__13^4*wit__2^4 - 24*(wit__0^2 + wit__11^2)*wit__12^2*wit__13^2*wit__2^2 - 24*(wit__0^2 + wit__11^2)*wit__13^2*wit__2^2*wit__4^2 - 5*wit__11^2*wit__2^2*wit__5^2)*wit__9^4 + 1/12*(12*wit__12^2*wit__13^4*wit__2^4 + 12*wit__13^4*wit__2^4*wit__4^2 + wit__12^2*wit__2^2*wit__5^2)*wit__9^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(2,5),List((0,"w_0","1","1"),(0,"abs_1","-rv_0*t_0 + r_0","1"),(8,"r_0","rv_0*t_0 - wit__6^2 + rp_0","-1"),(10,"rv_0","wit__9^2","1"),(0,"dhd_0","min_1","-1"),(0,"min_1","dhf_0 - max_1","1"),(1,"dhf_0","-wit__1^2 + max_1","-1"),(0,"max_1","wit__0^2","-1"),(0,"a_0","wit__2^2","1"),(0,"hp_0","wit__3^2","1"),(3,"rp_0","wit__6^2 + wit__8^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1/2","wit__2*wit__9^2*wit__5*wit__7"),("7/12","wit__2*wit__9*wit__5*wit__11*wit__12"),("1/12","-wit__2^3*wit__5*wit__13^2 + wit__2*wit__5*wit__12^2"),("1/12","wit__2^2*wit__9*wit__5*wit__11*wit__13"),("1","wit__2^2*wit__4*wit__6*wit__5"),("1/6","wit__2^2*wit__6*wit__5*wit__12")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"wit__13^6*wit__2^6 + wit__12^2*wit__13^4*wit__2^4 + wit__13^4*wit__2^4*wit__4^2 + 1/4*(4*wit_^2*wit__0^4 - 2*wit__0^6 + 4*wit__11^6 + 4*(wit_^2 + wit__0^2)*wit__11^4 - 4*(wit__0^2 + wit__11^2)*wit__2^2*wit__3^2 + 2*(4*wit_^2*wit__0^2 - wit__0^4)*wit__11^2 - 4*(h_0*wit__0^2 + h_0*wit__11^2)*wit__2^2 + (t_0*wit__2^2 + wit__0^2 + wit__11^2)*wit__7^2)*wit__9^6 - 1/2*(4*wit__13^2*wit__2^4*wit__3^2 + 4*h_0*wit__13^2*wit__2^4 + wit__2^2*wit__6^2*wit__7^2 - 2*(2*wit_^2*wit__0^2 - wit__0^4 + 3*wit__11^4 + 2*(wit_^2 + wit__0^2)*wit__11^2)*wit__13^2*wit__2^2 - 2*(wit__0^2*wit__11^2 + wit__11^4)*wit__12^2 - 2*(wit__0^2*wit__11^2 + wit__11^4)*wit__4^2)*wit__9^4 + 1/2*(2*wit__12^2*wit__13^2*wit__2^4 + 2*wit__13^2*wit__2^4*wit__4^2 + wit__12^4*wit__2^2)*wit__6^2 + 1/4*((wit__0^2 + wit__11^2)*wit__12^4 - 2*(t_0*wit__12^2*wit__13^2 - 2*(wit_^2 + wit__0^2 + 3*wit__11^2)*wit__13^4)*wit__2^4 - (t_0*wit__12^4 - 6*(wit__0^2 + wit__11^2)*wit__12^2*wit__13^2)*wit__2^2 - 2*(t_0*wit__13^2*wit__2^4 - 3*(wit__0^2 + wit__11^2)*wit__13^2*wit__2^2 - 2*(wit__0^2 + wit__11^2)*wit__12^2)*wit__4^2)*wit__9^2"),(1,"wit__12^2*wit__13^2*wit__2^4 + wit__13^2*wit__2^4*wit__4^2 - 1/2*wit__2^2*wit__7^2*wit__9^4 + 1/2*wit__12^4*wit__2^2"),(2,"-1/4*(wit__0^4 + 2*wit__2^2*wit__3^2 + 2*h_0*wit__2^2)*wit__9^8 + 1/2*(wit__0^2*wit__12^2 + wit__0^2*wit__4^2)*wit__9^6 - 1/4*(wit__12^4 + 2*wit__12^2*wit__4^2)*wit__9^4"),(3,"0"),(4,"-1/4*(t_0^3*wit__2^6 + 4*wit_^4*wit__0^2 - 4*wit_^2*wit__0^4 + 3*wit__0^6 - 2*wit__0^4*wit__1^2 + (4*wit_^2 - wit__0^2)*wit__11^4 + (4*t_0^2*wit_^2 - 3*t_0^2*wit__0^2 + t_0^2*wit__11^2 - 4*h_0*t_0)*wit__2^4 + (4*wit_^4 + wit__0^4)*wit__11^2 + (4*t_0*wit_^4 + t_0*wit__0^4 + t_0*wit__11^4 - 8*h_0*wit_^2 - 4*(t_0*wit_^2 - 2*h_0)*wit__0^2 - 4*h_0*wit__1^2 + 2*(2*t_0*wit_^2 - t_0*wit__0^2)*wit__11^2)*wit__2^2 - 4*(t_0*wit__2^4 + (2*wit_^2 - 2*wit__0^2 + wit__1^2)*wit__2^2)*wit__3^2 + (wit__0^2 - wit__1^2)*wit__7^2)*wit__9^8 - 1/4*(4*t_0^2*wit__13^2*wit__2^6 - 4*wit__13^2*wit__2^4*wit__3^2 + 4*(3*t_0*wit_^2 - 2*t_0*wit__0^2 + t_0*wit__11^2 - h_0)*wit__13^2*wit__2^4 + (8*wit_^4 - 4*wit_^2*wit__0^2 + 2*wit__0^4 + wit__11^4 + 4*(3*wit_^2 - wit__0^2)*wit__11^2)*wit__13^2*wit__2^2 - (4*wit__0^4 - 4*wit__0^2*wit__1^2 + 4*wit__0^2*wit__11^2 + 3*wit__11^4)*wit__12^2 - 2*(wit__0^4 - 2*wit__0^2*wit__1^2 + 2*wit__0^2*wit__11^2 + 2*wit__11^4 - 2*wit__2^2*wit__3^2 - 2*h_0*wit__2^2)*wit__4^2 - (wit__12^2 + wit__4^2)*wit__7^2)*wit__9^6 - 1/4*(6*t_0*wit__13^4*wit__2^6 + 3*(4*wit_^2 - 2*wit__0^2 + wit__11^2)*wit__13^4*wit__2^4 - 2*(3*wit__0^2 - wit__1^2 + 3*wit__11^2)*wit__12^2*wit__13^2*wit__2^2 + (wit__0^2 - wit__1^2 - wit__11^2)*wit__12^4 - 4*(wit__0^2 + wit__11^2)*wit__4^4 - 2*((3*wit__0^2 - wit__1^2 + 4*wit__11^2)*wit__13^2*wit__2^2 + 2*(wit__1^2 + wit__11^2)*wit__12^2)*wit__4^2)*wit__9^4 - 1/12*(9*wit__13^6*wit__2^6 - 9*wit__12^2*wit__13^4*wit__2^4 + 3*wit__12^4*wit__13^2*wit__2^2 - 6*wit__13^2*wit__2^2*wit__4^4 - 3*(4*wit__13^4*wit__2^4 + wit__12^4)*wit__4^2 - (wit__13^2*wit__2^4 + 7*wit__12^2*wit__2^2)*wit__5^2)*wit__9^2"),(5,"3/4*wit__13^6*wit__2^6 + 1/4*wit__12^2*wit__13^4*wit__2^4 + 1/4*wit__12^4*wit__13^2*wit__2^2 + 1/2*wit__12^2*wit__13^2*wit__2^2*wit__4^2 + 1/4*(4*wit_^2*wit__0^4 - 2*wit__0^6 + 8*wit_^2*wit__0^2*wit__11^2 + 3*wit__11^6 - 4*wit__0^2*wit__2^2*wit__3^2 + 4*(wit_^2 + wit__0^2)*wit__11^4 - 4*h_0*wit__0^2*wit__2^2 + (2*wit__0^2 + wit__11^2)*wit__7^2)*wit__9^6 - 1/4*(4*wit__13^2*wit__2^4*wit__3^2 + 4*h_0*wit__13^2*wit__2^4 - wit__11^4*wit__12^2 - (8*wit_^2*wit__0^2 - 2*wit__0^4 + 9*wit__11^4 + 8*(wit_^2 + wit__0^2)*wit__11^2)*wit__13^2*wit__2^2 - 2*(wit__0^4 + 2*wit__2^2*wit__3^2 + 2*h_0*wit__2^2)*wit__4^2 + (wit__2^2*wit__6^2 + wit__12^2 + 2*wit__4^2)*wit__7^2)*wit__9^4 - 1/12*(wit__13^2*wit__2^4 + 7*wit__12^2*wit__2^2)*wit__5^2 + 1/4*(2*wit__12^2*wit__13^2*wit__2^4 + 2*wit__13^2*wit__2^4*wit__4^2 + wit__12^4*wit__2^2)*wit__6^2 + 1/4*((4*wit_^2 + 4*wit__0^2 + 9*wit__11^2)*wit__13^4*wit__2^4 - 2*wit__11^2*wit__13^2*wit__2^2*wit__4^2 - 4*(wit__0^2 + wit__11^2)*wit__4^4)*wit__9^2"),(6,"-1/4*wit__13^4*wit__2^8*wit__6^2 + 1/4*wit__13^2*wit__2^8*wit__6^4 - 1/4*wit__2^8*wit__6^6 - 3/4*wit__12^2*wit__13^4*wit__2^6 - 1/4*wit__12^4*wit__13^2*wit__2^4 - wit__13^2*wit__2^4*wit__4^4 - 1/4*wit__12^6*wit__2^2 + 1/4*(4*t_0^3*wit__2^8 - 4*(3*t_0^2*wit__1^2 + 2*h_0*t_0)*wit__2^6 + 4*(2*t_0*wit__1^4 + 2*h_0*wit__1^2 + h_0*wit__11^2)*wit__2^4 + (2*wit__0^4*wit__11^2 - wit__11^6)*wit__2^2 - 4*(2*t_0*wit__2^6 - (2*wit__1^2 + wit__11^2)*wit__2^4)*wit__3^2)*wit__9^6 + 1/4*(6*t_0^2*wit__13^2*wit__2^8 - 12*t_0*wit__1^2*wit__13^2*wit__2^6 + (2*wit__0^4 + 4*wit__1^4 - 3*wit__11^4)*wit__13^2*wit__2^4 - (4*wit__0^2*wit__11^2 + 3*wit__11^4)*wit__12^2*wit__2^2 + 2*(2*wit__2^4*wit__3^2 + 2*h_0*wit__2^4 + (wit__0^4 - 2*wit__0^2*wit__11^2 - 2*wit__11^4)*wit__2^2)*wit__4^2 - 2*(3*t_0^2*wit__2^8 + 2*wit__1^4*wit__2^4 - 2*wit__2^6*wit__3^2 - 2*(3*t_0*wit__1^2 + h_0)*wit__2^6)*wit__6^2)*wit__9^4 - 1/2*(2*wit__13^4*wit__2^6 + 2*wit__12^2*wit__13^2*wit__2^4 + wit__12^4*wit__2^2)*wit__4^2 - 1/12*(wit__13^2*wit__2^6 + 9*wit__12^2*wit__2^4 + 12*wit__2^4*wit__4^2)*wit__5^2 + 1/4*(4*t_0*wit__13^4*wit__2^8 - (4*wit__1^2 + 3*wit__11^2)*wit__13^4*wit__2^6 - 2*(2*wit__0^2 + 3*wit__11^2)*wit__12^2*wit__13^2*wit__2^4 - wit__11^2*wit__12^4*wit__2^2 - 4*(wit__0^2 + wit__11^2)*wit__2^2*wit__4^4 + 4*(t_0*wit__2^8 - wit__1^2*wit__2^6)*wit__6^4 - 4*((wit__0^2 + 2*wit__11^2)*wit__13^2*wit__2^4 + (wit__0^2 + wit__11^2)*wit__12^2*wit__2^2)*wit__4^2 - 4*(t_0*wit__13^2*wit__2^8 - wit__1^2*wit__13^2*wit__2^6)*wit__6^2)*wit__9^2"),(7,"-(t_0^3*wit__2^6 - wit_^2*wit__0^4 + wit__0^6 - wit__11^6 - (wit_^2 + wit__0^2)*wit__11^4 + (t_0^2*wit_^2 - t_0^2*wit__0^2 - t_0^2*wit__1^2 - 2*h_0*t_0)*wit__2^4 - (2*wit_^2*wit__0^2 - wit__0^4)*wit__11^2 + 2*(h_0*wit__0^2 + h_0*wit__11^2)*wit__2^2 - 2*(t_0*wit__2^4 - (wit__0^2 + wit__11^2)*wit__2^2)*wit__3^2)*wit__9^8 - (3*t_0^2*wit__13^2*wit__2^6 + 2*(t_0*wit_^2 - t_0*wit__0^2 - t_0*wit__1^2)*wit__13^2*wit__2^4 - (2*wit_^2*wit__0^2 - wit__0^4 + 3*wit__11^4 + 2*(wit_^2 + wit__0^2)*wit__11^2)*wit__13^2*wit__2^2 - (wit__0^4 + 2*wit__0^2*wit__11^2 + wit__11^4)*wit__12^2 - (wit__0^4 + 2*wit__0^2*wit__11^2 + wit__11^4)*wit__4^2)*wit__9^6 - (3*t_0*wit__13^4*wit__2^6 - (2*wit__0^2 + wit__1^2 + 3*wit__11^2)*wit__13^4*wit__2^4 - 2*(wit__0^2 + wit__11^2)*wit__12^2*wit__13^2*wit__2^2 - 2*(wit__0^2 + wit__11^2)*wit__13^2*wit__2^2*wit__4^2)*wit__9^4 + (wit__12^2*wit__13^4*wit__2^4 + wit__13^4*wit__2^4*wit__4^2)*wit__9^2")).map (s => (s._1,s._2.asTerm))))),
      Some((List(2),List((0,"w_0","-1","1"),(0,"abs_1","rv_0*t_0 - r_0","1"),(4,"r_0","rv_0*t_0 - wit__2^2","1"),(10,"rv_0","wit__9^2","1"),(0,"dhd_0","-min_1","1"),(0,"min_1","-dhf_0 - max_1","1"),(1,"dhf_0","wit__1^2 - max_1","1"),(0,"max_1","wit__0^2","-1"),(0,"a_0","wit__3^2","1"),(0,"hp_0","wit__4^2","1"),(2,"rp_0","wit__2^2 + wit__7^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__2*wit__10"),("1","wit__2*wit__13")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"wit__2^2"),(4,"0"),(5,"0"),(6,"-wit__2^2"),(7,"0")).map (s => (s._1,s._2.asTerm))))),
      Some((List(2),List((0,"w_0","1","1"),(0,"abs_1","rv_0*t_0 - r_0","1"),(4,"r_0","rv_0*t_0 - wit__2^2","1"),(10,"rv_0","wit__9^2","1"),(0,"dhd_0","min_1","-1"),(0,"min_1","dhf_0 - max_1","1"),(1,"dhf_0","-wit__1^2 + max_1","-1"),(0,"max_1","wit__0^2","-1"),(0,"a_0","wit__3^2","1"),(0,"hp_0","wit__4^2","1"),(2,"rp_0","wit__2^2 + wit__7^2","1")).map( s => (s._1,s._2.asTerm,s._3.asTerm,s._4.asTerm)),List(("1","wit__2*wit__10"),("1","wit__2*wit__13")).map( s => (s._1.asTerm,s._2.asTerm)),Some(List((0,"0"),(1,"0"),(2,"0"),(3,"wit__2^2"),(4,"0"),(5,"0"),(6,"-wit__2^2"),(7,"0")).map (s => (s._1,s._2.asTerm))))))

    val tactic = implyR(1) & DebuggingTactics.print("Before normalizing") & prop &
      DebuggingTactics.print("Before split") &
      OnAll(prepareArith2) &
      printGoal

    val res = proveBy(i, tactic)

    val result = lwits.zipWithIndex.map(
      prind =>
      {
        val _ = println("SOLVING",prind._2)
        prind._1 match
        {
          case Some(w) =>
            proveBy(res.subgoals(prind._2), linearElim(w._2) & genWitnessTac(w._1,w._3,w._4) & done)
          case None =>
            proveBy(res.subgoals(prind._2), nil)
        }
      }
    )

    print(result)
  }

}
